# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
from forte.utils import create_import_error_msg
from forte.common.configuration import Config
from forte.models.ner.conditional_random_field import ConditionalRandomField

try:
    import torch
    import torch.nn.functional as F
    import torch.nn.utils.rnn as rnn_utils
    from torch import nn
except ImportError as e1:
    raise ImportError(
        create_import_error_msg("torch", "models", "model factory")
    ) from e1


class BiRecurrentConvCRF(nn.Module):
    def __init__(
        self,
        word_vocab: Dict,
        char_vocab_size: int,
        tag_vocab_size: int,
        config_model: Config,
    ):
        super().__init__()
        try:
            import texar.torch as texar  # pylint: disable=import-outside-toplevel
            from texar.torch.modules.embedders import (  # pylint: disable=import-outside-toplevel
                WordEmbedder,
            )
        except ImportError as e2:
            raise ImportError(
                create_import_error_msg(
                    "texar-pytorch", "models", "BiRecurrentConvCRF"
                )
            ) from e2
        self.word_embedder = WordEmbedder(
            init_value=texar.data.Embedding(
                vocab=word_vocab,
                hparams={
                    "dim": config_model.word_emb.dim,
                    "file": config_model.embedding_path,
                    "read_fn": "load_glove",
                },
            ).word_vecs
        )

        self.char_embedder = WordEmbedder(
            vocab_size=char_vocab_size, hparams=config_model.char_emb
        )

        self.char_cnn = torch.nn.Conv1d(**config_model.char_cnn_conv)

        self.dropout_in = nn.Dropout2d(config_model.dropout_rate)
        # standard dropout
        self.dropout_rnn_in = nn.Dropout(config_model.dropout_rate)
        self.dropout_out = nn.Dropout(config_model.dropout_rate)

        self.rnn = nn.LSTM(
            config_model.bilstm_sentence_encoder.rnn_cell_fw.input_size,
            config_model.bilstm_sentence_encoder.rnn_cell_fw.kwargs.num_units,
            num_layers=1,
            batch_first=True,
            bidirectional=True,
        )

        self.dense = nn.Linear(
            config_model.bilstm_sentence_encoder.rnn_cell_fw.kwargs.num_units
            * 2,
            config_model.output_hidden_size,
        )

        self.tag_projection_layer = nn.Linear(
            config_model.output_hidden_size, tag_vocab_size
        )

        self.crf = ConditionalRandomField(
            tag_vocab_size, constraints=None, include_start_end_transitions=True
        )

        if config_model.initializer is None or callable(
            config_model.initializer
        ):
            self.initializer = config_model.initializer
        else:
            self.initializer = texar.core.layers.get_initializer(
                config_model["initializer"]
            )

        self.reset_parameters()

    def reset_parameters(self):
        if self.initializer is None:
            return

        for name, parameter in self.named_parameters():
            if name.find("embedder") == -1 and name.find("crf") == -1:
                if parameter.dim() == 1:
                    nn.init.constant_(parameter, 0.0)
                else:
                    self.initializer(parameter)

    def forward(self, input_word, input_char, target=None, mask=None, hx=None):
        """

        Args:
            input_word:
            input_char:
            target:
            mask:
            hx:

        Returns: the loss value

        """
        output, _, mask, _ = self.encode(input_word, input_char, mask, hx)

        logits = self.tag_projection_layer(output)
        log_likelihood = (
            self.crf.forward(logits, target, mask) / target.size()[0]
        )
        return -log_likelihood

    def decode(self, input_word, input_char, mask=None, hx=None):
        """
        Args:
            input_word:
            input_char:
            mask:
            hx:

        Returns:

        """
        output, _, mask, _ = self.encode(
            input_word, input_char, mask=mask, hx=hx
        )

        logits = self.tag_projection_layer(output)
        best_paths = self.crf.viterbi_tags(logits, mask.long())
        predicted_tags = [x for x, y in best_paths]
        predicted_tags = [torch.tensor(x).unsqueeze(0) for x in predicted_tags]
        try:
            import texar.torch as texar  # pylint: disable=import-outside-toplevel
        except ImportError as e3:
            raise ImportError(
                create_import_error_msg(
                    "texar-pytorch", "models", "BiRecurrentConvCRF"
                )
            ) from e3

        predicted_tags = texar.utils.pad_and_concat(
            predicted_tags, axis=0, pad_constant_values=0
        )

        return predicted_tags

    def encode(self, input_word, input_char, mask=None, hx=None):
        # output from rnn [batch, length, tag_space]
        length = mask.sum(dim=1).long()

        # [batch, length, word_dim]
        word = self.word_embedder(input_word)
        word = self.dropout_in(word)

        # [batch, length, char_length, char_dim]
        char = self.char_embedder(input_char)
        char_size = char.size()
        # first transform to [batch * length, char_length, char_dim]
        # then transpose to [batch * length, char_dim, char_length]
        char = char.view(
            char_size[0] * char_size[1], char_size[2], char_size[3]
        ).transpose(1, 2)
        # put into cnn [batch*length, char_filters, char_length]
        # then put into maxpooling [batch * length, char_filters]
        char, _ = self.char_cnn(char).max(dim=2)
        # reshape to [batch, length, char_filters]
        char = torch.tanh(char).view(char_size[0], char_size[1], -1)

        # independently apply dropout to word and characters
        char = self.dropout_in(char)

        # concatenate word and char [batch, length, word_dim+char_filter]
        input = torch.cat([word, char], dim=2)
        input = self.dropout_rnn_in(input)

        # prepare packed_sequence
        seq_input, hx, rev_order, mask = prepare_rnn_seq(
            input, length, hx=hx, masks=mask, batch_first=True
        )
        self.rnn.flatten_parameters()
        seq_output, hn = self.rnn(seq_input, hx=hx)
        output, hn = recover_rnn_seq(
            seq_output, rev_order, hx=hn, batch_first=True
        )

        # apply dropout for the output of rnn
        output = self.dropout_out(output)

        # [batch, length, tag_space]
        output = self.dropout_out(F.elu(self.dense(output)))

        return output, hn, mask, length


def prepare_rnn_seq(rnn_input, lengths, hx=None, masks=None, batch_first=False):
    """

    Args:
        rnn_input: [seq_len, batch, input_size]:
            tensor containing the features of the input sequence.
        lengths: [batch]:
            tensor containing the lengthes of the input sequence
        hx: [num_layers * num_directions, batch, hidden_size]:
            tensor containing the initial hidden state for each element
            in the batch.
        masks: [seq_len, batch]:
            tensor containing the mask for each element in the batch.
        batch_first:
            If True, then the input and output tensors are provided as
            [batch, seq_len, feature].

    Returns:

    """

    def check_decreasing(lengths):
        lens, order = torch.sort(lengths, dim=0, descending=True)
        if torch.ne(lens, lengths).sum() == 0:
            return None
        else:
            _, rev_order = torch.sort(order)
            return lens, order, rev_order

    check_res = check_decreasing(lengths)

    if check_res is None:
        lens = lengths
        rev_order = None
    else:
        lens, order, rev_order = check_res
        batch_dim = 0 if batch_first else 1
        rnn_input = rnn_input.index_select(batch_dim, order)
        if hx is not None:
            if isinstance(hx, tuple):
                hx, cx = hx
                hx = hx.index_select(1, order)
                cx = cx.index_select(1, order)
                hx = (hx, cx)
            else:
                hx = hx.index_select(1, order)

    lens = lens.tolist()
    seq = rnn_utils.pack_padded_sequence(
        rnn_input, lens, batch_first=batch_first
    )

    if masks is not None:
        if batch_first:
            masks = masks[:, : lens[0]]
        else:
            masks = masks[: lens[0]]

    return seq, hx, rev_order, masks


def recover_rnn_seq(seq, rev_order, hx=None, batch_first=False):
    output, _ = rnn_utils.pad_packed_sequence(seq, batch_first=batch_first)
    if rev_order is not None:
        batch_dim = 0 if batch_first else 1
        output = output.index_select(batch_dim, rev_order)
        if hx is not None:
            # hack lstm
            if isinstance(hx, tuple):
                hx, cx = hx
                hx = hx.index_select(1, rev_order)
                cx = cx.index_select(1, rev_order)
                hx = (hx, cx)
            else:
                hx = hx.index_select(1, rev_order)
    return output, hx



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.data_pack import DataPack
from forte.data.ontology.top import Annotation
from forte.processors.base import PackProcessor
from forte.utils.utils import get_class
from forte.utils import create_import_error_msg


__all__ = [
    "PretrainedEncoder",
]


class PretrainedEncoder(PackProcessor):
    r"""A wrapper of Texar pre-trained encoders.

    This processor will compute the embedding vectors for entries of type
    :class:`~forte.data.ontology.top.Annotation` using pre-trained models. The user can specify the
    pre-trained model type and the annotation class name via configuration.
    For the full list of pre-trained models supported, see
    :meth:`default_config` for more details. The processor will add embedding
    vector for all entries matching the specified entry type. The resulting
    vector can be accessed by the embedding field of the annotations.
    """

    def __init__(self):
        super().__init__()
        self.tokenizer = None
        self.encoder = None
        self.entry_type = None
        try:
            import texar.torch as tx  # pylint: disable=import-outside-toplevel
        except ImportError as err:
            raise ImportError(
                create_import_error_msg("texar-pytorch", "nlp", "NLP support")
            ) from err
        self.name2tokenizer = {
            "bert-base-uncased": tx.data.BERTTokenizer,
            "bert-large-uncased": tx.data.BERTTokenizer,
            "bert-base-cased": tx.data.BERTTokenizer,
            "bert-large-cased": tx.data.BERTTokenizer,
            "bert-base-multilingual-uncased": tx.data.BERTTokenizer,
            "bert-base-multilingual-cased": tx.data.BERTTokenizer,
            "bert-base-chinese": tx.data.BERTTokenizer,
            "biobert-v1.0-pmc": tx.data.BERTTokenizer,
            "biobert-v1.0-pubmed-pmc": tx.data.BERTTokenizer,
            "biobert-v1.0-pubmed": tx.data.BERTTokenizer,
            "biobert-v1.1-pubmed": tx.data.BERTTokenizer,
            "scibert-scivocab-uncased": tx.data.BERTTokenizer,
            "scibert-scivocab-cased": tx.data.BERTTokenizer,
            "scibert-basevocab-uncased": tx.data.BERTTokenizer,
            "scibert-basevocab-cased": tx.data.BERTTokenizer,
            "gpt2-small": tx.data.GPT2Tokenizer,
            "gpt2-medium": tx.data.GPT2Tokenizer,
            "gpt2-large": tx.data.GPT2Tokenizer,
            "gpt2-xl": tx.data.GPT2Tokenizer,
            "roberta-base": tx.data.RoBERTaTokenizer,
            "roberta-large": tx.data.RoBERTaTokenizer,
            "T5-Small": tx.data.T5Tokenizer,
            "T5-Base": tx.data.T5Tokenizer,
            "T5-Large": tx.data.T5Tokenizer,
            "T5-3B": tx.data.T5Tokenizer,
            "T5-11B": tx.data.T5Tokenizer,
            "xlnet-based-cased": tx.data.XLNetTokenizer,
            "xlnet-large-cased": tx.data.XLNetTokenizer,
        }
        self.name2encoder = {
            "bert-base-uncased": tx.modules.BERTEncoder,
            "bert-large-uncased": tx.modules.BERTEncoder,
            "bert-base-cased": tx.modules.BERTEncoder,
            "bert-large-cased": tx.modules.BERTEncoder,
            "bert-base-multilingual-uncased": tx.modules.BERTEncoder,
            "bert-base-multilingual-cased": tx.modules.BERTEncoder,
            "bert-base-chinese": tx.modules.BERTEncoder,
            "biobert-v1.0-pmc": tx.modules.BERTEncoder,
            "biobert-v1.0-pubmed-pmc": tx.modules.BERTEncoder,
            "biobert-v1.0-pubmed": tx.modules.BERTEncoder,
            "biobert-v1.1-pubmed": tx.modules.BERTEncoder,
            "scibert-scivocab-uncased": tx.modules.BERTEncoder,
            "scibert-scivocab-cased": tx.modules.BERTEncoder,
            "scibert-basevocab-uncased": tx.modules.BERTEncoder,
            "scibert-basevocab-cased": tx.modules.BERTEncoder,
            "gpt2-small": tx.modules.GPT2Encoder,
            "gpt2-medium": tx.modules.GPT2Encoder,
            "gpt2-large": tx.modules.GPT2Encoder,
            "gpt2-xl": tx.modules.GPT2Encoder,
            "roberta-base": tx.modules.RoBERTaEncoder,
            "roberta-large": tx.modules.RoBERTaEncoder,
            "T5-Small": tx.modules.T5Encoder,
            "T5-Base": tx.modules.T5Encoder,
            "T5-Large": tx.modules.T5Encoder,
            "T5-3B": tx.modules.T5Encoder,
            "T5-11B": tx.modules.T5Encoder,
            "xlnet-based-cased": tx.modules.XLNetEncoder,
            "xlnet-large-cased": tx.modules.XLNetEncoder,
        }

    def available_checkpoints(self):
        return list(self.name2tokenizer.keys())

    # pylint: disable=unused-argument
    def initialize(self, resources: Resources, configs: Config):
        if configs.pretrained_model_name in self.name2tokenizer:
            self.tokenizer = self.name2tokenizer[configs.pretrained_model_name](
                pretrained_model_name=configs.pretrained_model_name
            )
            self.encoder = self.name2encoder[configs.pretrained_model_name](
                pretrained_model_name=configs.pretrained_model_name
            )
        else:
            raise ValueError("Unrecognized pre-trained model name.")

        self.entry_type = get_class(configs.entry_type)
        if not isinstance(self.entry_type, Annotation) and not issubclass(
            self.entry_type, Annotation
        ):
            raise ValueError("entry_type must be annotation type.")

    def _process(self, input_pack: DataPack):
        try:
            import torch  # pylint: disable=import-outside-toplevel
        except ImportError as err:
            raise ImportError(
                create_import_error_msg("texar-pytorch", "nlp", "NLP support")
            ) from err
        for entry in input_pack.get(entry_type=self.entry_type):  # type: ignore
            input_ids, segment_ids, _ = self.tokenizer.encode_text(
                text_a=entry.text
            )

            input_ids = torch.tensor([input_ids])
            segment_ids = torch.tensor([segment_ids])
            input_length = (1 - (input_ids == 0).int()).sum(dim=1)

            output, _ = self.encoder(input_ids, input_length, segment_ids)
            entry.embedding = output.tolist()

    @classmethod
    def default_configs(cls):
        r"""This default configurations for :class:`PretrainedEncoder`.

        Here:

        `"pretrained_model_name"`: str or None
            The name of the pre-trained models including:

            * **Standard BERT**: proposed in (`Devlin et al`. 2018)
              `BERT: Pre-training of Deep Bidirectional Transformers for
              Language Understanding`_
              . A bidirectional Transformer language model pre-trained on large
              text corpora. Available model names include:

                * ``bert-base-uncased``: 12-layer, 768-hidden, 12-heads,
                  110M parameters.
                * ``bert-large-uncased``: 24-layer, 1024-hidden, 16-heads,
                  340M parameters.
                * ``bert-base-cased``: 12-layer, 768-hidden, 12-heads ,
                  110M parameters.
                * ``bert-large-cased``: 24-layer, 1024-hidden, 16-heads,
                  340M parameters.
                * ``bert-base-multilingual-uncased``: 102 languages, 12-layer,
                  768-hidden, 12-heads, 110M parameters.
                * ``bert-base-multilingual-cased``: 104 languages, 12-layer,
                  768-hidden, 12-heads, 110M parameters.
                * ``bert-base-chinese``: Chinese Simplified and Traditional,
                  12-layer, 768-hidden, 12-heads, 110M parameters.

            * **BioBERT**: proposed in (`Lee et al`. 2019)
              `BioBERT: a pre-trained biomedical language representation model
              for biomedical text mining`_
              . A domain specific language representation model pre-trained on
              large-scale biomedical corpora. Based on the BERT architecture,
              BioBERT effectively transfers the knowledge from a large amount
              of biomedical texts to biomedical text mining models with minimal
              task-specific architecture modifications. Available model names
              include:

              * ``biobert-v1.0-pmc``: BioBERT v1.0 (+ PMC 270K) - based on
                BERT-base-Cased (same vocabulary).
              * ``biobert-v1.0-pubmed-pmc``: BioBERT v1.0 (+ PubMed 200K + PMC
                270K) - based on BERT-base-Cased (same vocabulary).
              * ``biobert-v1.0-pubmed``: BioBERT v1.0 (+ PubMed 200K) - based on
                BERT-base-Cased (same vocabulary).
              * ``biobert-v1.1-pubmed``: BioBERT v1.1 (+ PubMed 1M) - based on
                BERT-base-Cased (same vocabulary).

            * **SciBERT**: proposed in (`Beltagy et al`. 2019)
              `SciBERT: A Pretrained Language Model for Scientific Text`_. A
              BERT model trained on scientific text. SciBERT leverages
              unsupervised pre-training on a large multi-domain corpus of
              scientific publications to improve performance on downstream
              scientific NLP tasks. Available model names include:

              * ``scibert-scivocab-uncased``: Uncased version of the model
                trained on its own vocabulary.
              * ``scibert-scivocab-cased``: Cased version of the model trained
                on its own vocabulary.
              * ``scibert-basevocab-uncased``: Uncased version of the model
                trained on the original BERT vocabulary.
              * ``scibert-basevocab-cased``: Cased version of the model trained
                on the original BERT vocabulary.

            * The GPT2 model was proposed in
              `Language Models are Unsupervised Multitask Learners`_
              by `Radford et al.` from OpenAI. It is a unidirectional
              Transformer model pre-trained using the vanilla language modeling
              objective on a large corpus.

              The available GPT2 models are as follows:

              * ``gpt2-small``: Small version of GPT-2, 124M parameters.
              * ``gpt2-medium``: Medium version of GPT-2, 355M parameters.
              * ``gpt2-large``: Large version of GPT-2, 774M parameters.
              * ``gpt2-xl``: XL version of GPT-2, 1558M parameters.

            * The RoBERTa model was proposed in (`Liu et al`. 2019)
              `RoBERTa: A Robustly Optimized BERT Pretraining Approach`_.
              As a variant of the standard BERT model, RoBERTa trains for more
              iterations on more data with a larger batch size as well as other
              tweaks in pre-training. Differing from the standard BERT, the
              RoBERTa model does not use segmentation embedding. Available
              model names include:

              * ``roberta-base``: RoBERTa using the BERT-base architecture,
                125M parameters.
              * ``roberta-large``: RoBERTa using the BERT-large architecture,
                355M parameters.

            * The T5 model treats multiple NLP tasks in a similar manner by
              encoding the different tasks as text directives in the input
              stream. This enables a single model to be trained supervised on a
              wide variety of NLP tasks.

              The T5 model examines factors relevant for leveraging transfer
              learning at scale from pure unsupervised pre-training to
              supervised tasks. It is discussed in much detail in
              `Exploring the Limits of Transfer Learning with a Unified
              Text-to-Text Transformer`_
              from Google.

              The available T5 models are as follows:

              * ``T5-Small``: Small version of T5, 60 million parameters.
              * ``T5-Base``: Base-line version of T5, 220 million parameters.
              * ``T5-Large``: Large Version of T5, 770 million parameters.
              * ``T5-3B``: A version of T5 with 3 billion parameters.
              * ``T5-11B``: A version of T5 with 11 billion parameters.

            * The XLNet model was proposed in
              `XLNet: Generalized Autoregressive Pretraining for Language
              Understanding`_
              by `Yang et al.` It is based on the Transformer-XL model,
              pre-trained on a large corpus using a language modeling objective
              that considers all permutations of the input sentence.

              The available XLNet models are as follows:

              * ``xlnet-based-cased``: 12-layer, 768-hidden, 12-heads. This
                model is trained on full data (different from the one in the
                paper).
              * ``xlnet-large-cased``: 24-layer, 1024-hidden, 16-heads.

        `"entry_type"`: str
            The annotation type that user want to generate embbeding on .

        .. _`BERT: Pre-training of Deep Bidirectional Transformers for Language
         Understanding`:
            https://arxiv.org/abs/1810.04805

        .. _`BioBERT: a pre-trained biomedical language representation model
        for biomedical text mining`:
            https://arxiv.org/abs/1901.08746

        .. _`SciBERT: A Pretrained Language Model for Scientific Text`:
            https://arxiv.org/abs/1903.10676

        .. _`Language Models are Unsupervised Multitask Learners`:
            https://openai.com/blog/better-language-models/

        .. _`RoBERTa: A Robustly Optimized BERT Pretraining Approach`:
            https://arxiv.org/abs/1907.11692

        .. _`Exploring the Limits of Transfer Learning with a Unified
        Text-to-Text Transformer`:
            https://arxiv.org/abs/1910.10683

        .. _`XLNet: Generalized Autoregressive Pretraining for Language
        Understanding`:
            http://arxiv.org/abs/1906.08237
        """

        return {
            "pretrained_model_name": "bert-base-uncased",
            "entry_type": "ft.onto.base_ontology.Sentence",
        }



# Copyright 2019-2021 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC
from typing import Dict, Any, Optional, Union

import yaml

from forte.common import ProcessorConfigError
from forte.common.configuration import Config
from forte.utils import get_full_module_name

__all__ = ["Configurable"]


class Configurable(ABC):
    """
    Classes that implement the Configurable interface enables the classes to
    easily manage configurations. The extended classes should implement
    `default_configs` to provide additional default configurations. A
    `make_configs` functions is provided to create the configurations. There
    are the following behaviors:

      - `default_configs`: the default configs provide the configurations
        structure accepted by this class. This behavior is the same as
        :class:`texar.torch.HParams`. If the current class extends another
        `Configurable` class, the ``default_configs`` of the parent class will
        be inherited in an recursive way. If the same configuration key is
        defined for a parent and child class, the one from the child class
        will prevail.
      - `make_configs`: this function accepts user configurations and create
        the configuration by merging the user config and the default configs.
        Check the docstring of `make_configs` for its behaviors.
    """

    @classmethod
    def _default_configs(cls) -> Config:
        # pylint: disable=protected-access
        merged = Config(cls.default_configs(), {}, allow_new_hparam=True)
        for base in cls.__bases__:
            if hasattr(base, "_default_configs"):
                merged = Config(
                    merged,
                    base._default_configs().todict(),  # type: ignore
                    allow_new_hparam=True,
                )
                break
        return merged

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        return {}

    @classmethod
    def make_configs(
        cls,
        configs: Optional[Union[Config, Dict[str, Any]]],
    ) -> Config:
        """
        Create the configuration by merging the
        provided config with the `default_configs`.

        The following config conventions are expected:

          - The top level key can be a special `@config_path`.

          - `@config_path` should be point to a file system path, which will
            be a YAML file containing configurations.

          - Other key values in the configs will be considered as parameters.

        Args:
            configs: The input config to be merged with the default config.

        Returns:
            The merged configuration.
        """
        merged_configs: Dict = {}

        if configs is not None:
            if isinstance(configs, Config):
                configs = configs.todict()

            if configs.get("@config_path", None) is not None:
                with open(configs.pop("@config_path"), encoding="utf-8") as f:
                    filebased_configs = yaml.safe_load(f)
            else:
                filebased_configs = {}

            merged_configs.update(filebased_configs)

            merged_configs.update(configs)

        try:
            final_configs = Config(
                merged_configs, cls._default_configs().todict()
            )
        except ValueError as e:
            raise ProcessorConfigError(
                f"Configuration error for the processor "
                f"{get_full_module_name(cls)}."
            ) from e

        return final_configs



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Index processor
"""
from abc import ABC
from typing import Dict, Any, List, Tuple

from forte.common import Resources
from forte.common.configuration import Config
from forte.data.data_pack import DataPack
from forte.processors.base.pack_processor import PackProcessor

__all__ = ["IndexProcessor", "IndexProcessorWithDatapack"]


class IndexProcessor(PackProcessor, ABC):
    r"""A  base processor for indexing documents into traditional indexers like
    Elasticsearch and/or dense vector indexers like Faiss. Subclasses need to
    implement :meth:`IndexProcessor::_bulk_process`.

    """

    # pylint: disable=useless-super-delegation
    def __init__(self) -> None:
        super().__init__()
        self.documents: List[Dict[str, str]] = []

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        return {"batch_size": 128}

    def _bulk_process(self):
        r"""Subclasses of :class:`IndexProcessor` should implement this method
        to bulk add the documents into the index.
        """
        raise NotImplementedError

    def _field_names(self) -> List[str]:
        r"""Subclasses of :class:`IndexProcessor` should implement this method
          to provide the field name for indexing.
          The return value from :func:`_content_for_index` will be added into
          these fields. The length of the return value of this function should
          be the same as the return value for :func:`_content_for_index`.
        Returns:

        """
        raise NotImplementedError

    def _content_for_index(self, input_pack: DataPack) -> List[str]:
        raise NotImplementedError

    def _process(self, input_pack: DataPack):
        # self.documents.append((str(input_pack.pack_id), input_pack.text))
        index_pairs: Dict[str, str] = dict(
            zip(self._field_names(), self._content_for_index(input_pack))
        )
        self.documents.append(index_pairs)

        if len(self.documents) == self.configs.batch_size:
            self._bulk_process()
            self.documents = []

    def flush(self):
        if len(self.documents) > 0:
            self._bulk_process()


class IndexProcessorWithDatapack(PackProcessor, ABC):
    r"""A  base processor for indexing a document with its original datapack
    into traditional indexers like Elasticsearch.
    Subclasses need to implement
    :meth:`IndexProcessorWithDatapack::_bulk_process`.
    """

    # pylint: disable=useless-super-delegation
    def __init__(self) -> None:
        super().__init__()
        self.documents: List[Tuple[str, str, str]] = []

    # pylint: disable=attribute-defined-outside-init
    def initialize(self, resources: Resources, configs: Config):
        self.resources = resources
        self.config = configs

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        config = super().default_configs()
        config.update({"batch_size": 128})
        return config

    def _bulk_process(self):
        r"""Subclasses of :class:`IndexProcessorWithDatapack`
        should implement this method
        to bulk add the documents into the index.
        """
        raise NotImplementedError

    def _process(self, input_pack: DataPack):
        serialized_datapack: str = input_pack.to_string()

        self.documents.append(
            (str(input_pack.pack_id), input_pack.text, serialized_datapack)
        )

        if len(self.documents) == self.config.batch_size:
            self._bulk_process()
            self.documents = []

    def flush(self):
        if len(self.documents) > 0:
            self._bulk_process()



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=attribute-defined-outside-init
from typing import Dict, Any

from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.multi_pack import MultiPack
from forte.data.ontology.top import Query
from forte.processors.base import MultiPackProcessor
from forte.utils import create_class_with_kwargs
from ft.onto.base_ontology import Document

__all__ = ["SearchProcessor"]


class SearchProcessor(MultiPackProcessor):
    r"""This processor searches for relevant documents for a query"""

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)
        # Replace explicit class with configuration class name.
        self.index = create_class_with_kwargs(
            self.configs.indexer_class,
            class_args={"config": self.configs.indexer_configs},
        )
        self.index.load(self.configs.model_dir)
        self.k = self.configs.k or 5

    def _process(self, input_pack: MultiPack):
        query_pack = input_pack.get_pack(self.configs.query_pack_name)
        first_query = list(query_pack.get(Query))[0]
        results = self.index.search(first_query.value, self.k)
        documents = [r[1] for result in results for r in result]

        packs = {}
        for i, doc in enumerate(documents):
            pack = input_pack.add_pack()
            pack.set_text(doc)

            Document(pack, 0, len(doc))
            packs[self.configs.response_pack_name_prefix + f"_{i}"] = pack

        input_pack.update_pack(packs)

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        return {
            "model_dir": None,
            "query_pack_name": "query",
            "response_pack_name_prefix": "doc",
            "indexer_class": "fortex.faiss.embedding_based_indexer"
            ".EmbeddingBasedIndexer",
            "indexer_configs": {
                "index_type": "GpuIndexFlatIP",
                "dim": 768,
                "device": "gpu0",
            },
        }



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from forte.data.readers.conll03_reader import *
from forte.data.readers.conllu_ud_reader import *
from forte.data.readers.ms_marco_passage_reader import *
from forte.data.readers.deserialize_reader import *
from forte.data.readers.html_reader import *
from forte.data.readers.multipack_sentence_reader import *
from forte.data.readers.multipack_terminal_reader import *
from forte.data.readers.ontonotes_reader import *
from forte.data.readers.plaintext_reader import *
from forte.data.readers.prodigy_reader import *
from forte.data.readers.race_multi_choice_qa_reader import *
from forte.data.readers.string_reader import *
from forte.data.readers.sem_eval_task8_reader import *
from forte.data.readers.sst2_reader import *
from forte.data.readers.openie_reader import *
from forte.data.readers.ag_news_reader import *
from forte.data.readers.largemovie_reader import *
from forte.data.readers.misc_readers import *
from forte.data.readers.classification_reader import *
from forte.data.readers.audio_reader import *



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Forte defined exceptions.
"""

__all__ = [
    "PackIndexError",
    "IncompleteEntryError",
    "EntryNotFoundError",
    "ProcessorConfigError",
    "PackDataException",
    "ProcessFlowException",
    "ProcessExecutionException",
    "ValidationError",
    "ResourceError",
    "ExpectedRecordNotFound",
    "InvalidOperationException",
    "UnknownOntologyClassException",
]


class PackIndexError(Exception):
    r"""Raise this error when there is a problem accessing the indexes pack
    data.
    """
    pass


class IncompleteEntryError(Exception):
    r"""Raise this error when the entry is not complete."""
    pass


class EntryNotFoundError(ValueError):
    r"""Raise this error when the entry is not found in the data pack."""
    pass


class ProcessorConfigError(ValueError):
    r"""Raise this error when the there is a problem with the processor
    configuration.
    """
    pass


class ResourceError(ValueError):
    r"""Raise this error when there are problems accessing Resources of
    the processor.
    """
    pass


class PackDataException(Exception):
    r"""Raise this error when the data in pack is wrong."""
    pass


class ProcessFlowException(Exception):
    r"""Raise this when errors happen in flow control"""
    pass


class ProcessExecutionException(Exception):
    r"""Raise this when process execution fail"""
    pass


class ValidationError(Exception):
    r"""Raise this error when input validate fail"""
    pass


class ExpectedRecordNotFound(ValueError):
    r"""Raise this error when expected type for input is not in input record"""
    pass


class InvalidOperationException(Exception):
    """Throw when a method call is invalid for the object's current state"""

    pass


class UnknownOntologyClassException(Exception):
    """Throw when there are unrecognized ontology class in the input"""

    pass



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import json
from typing import Tuple, Any, Dict, Union


from forte.data.ontology import Annotation
from forte.common.configuration import Config
from forte.processors.data_augment.algorithms.single_annotation_op import (
    SingleAnnotationAugmentOp,
)
from forte.utils import create_import_error_msg

__all__ = ["CharacterFlipOp"]


class CharacterFlipOp(SingleAnnotationAugmentOp):
    r"""
    A uniform generator that randomly flips a character with a similar
    looking character from a predefined dictionary imported from
    `"https://github.com/facebookresearch/AugLy/blob/main/" +
    "augly/text/augmenters/utils.py"`.
    (For example: the cat drank milk -> t/-/3 c@t d12@nk m!|_1<).

    Args:
        string: input string whose characters need to be replaced,

        dict_path: the `url` or the path to the pre-defined
                    typo json file,

        configs: prob(float): The probability of replacement,
                    should fall in [0, 1].
    """

    def __init__(self, configs: Union[Config, Dict[str, Any]]) -> None:
        try:
            import requests  # pylint: disable=import-outside-toplevel
        except ImportError as e:
            raise ImportError(
                create_import_error_msg(
                    "requests", "data_aug", "data augment support"
                )
            ) from e
        super().__init__(configs)

        self.dict_path = self.configs["dict_path"]

        try:
            r = requests.get(self.dict_path)
            self.data = r.json()
        except requests.exceptions.RequestException:
            with open(self.dict_path, encoding="utf8") as json_file:
                self.data = json.load(json_file)

    def _flip(self, char: str):
        r"""
        Flips character with similar character from input dictionary.

        Args:
            char: input character.
        Returns:
            the modified character.
        """
        if char in self.data:
            return random.choice(self.data[char])
        else:
            return char

    def single_annotation_augment(
        self, input_anno: Annotation
    ) -> Tuple[bool, str]:
        r"""
        Takes in the annotated string and performs the character
        flip operation on it that randomly augments few characters
        from it based on the probability value in the configs.

        Args:
            input_anno: the input annotation.
        Returns:
            A tuple with the first element being a boolean value indicating
            whether the replacement happens, and the second element is the
            final augmented string.
        """
        augmented_string = ""
        for char in input_anno.text:
            if char == " " or random.random() > self.configs["prob"]:
                augmented_string += char
            else:
                augmented_string += self._flip(char)
        return True, augmented_string

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        """
        Returns:
            A dictionary with the default config for this processor.
        Following are the keys for this dictionary:
            - dict_path (str):
                the `url` or the path to the pre-defined
                typo `json` file. One example dictionary is provided
                at `https://raw.githubusercontent.com/ArnavParekhji`
                `/temporaryJson/main/character_flip.json`
                as a default value.
            - prob (float):
                The probability of replacement. This value
                should fall in [0, 1]. Default value is 0.1
        """
        return {
            "dict_path": (
                "https://raw.githubusercontent.com/ArnavParekhji/"
                + "temporaryJson/main/character_flip.json"
            ),
            "prob": 0.1,
        }



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
__all__ = ["Vocabulary", "VocabFilter", "FrequencyVocabFilter"]
from abc import ABC
from collections import Counter
from typing import List, Tuple, Dict, Union, Hashable, Iterable, Optional
from typing import TypeVar, Generic, Any, Set
from asyml_utilities.special_tokens import SpecialTokens

from forte.common import InvalidOperationException

ElementType = TypeVar("ElementType", bound=Hashable)


class Vocabulary(Generic[ElementType]):
    r"""This class will store "Elements" that are added, assign "Ids" to them
    and return "Representations" if queried. These three are the main concepts
    in this class.

    1. Element: Any hash-able instance that the user want to store.
    2. Id: Each element will have an unique Id, which is an integer.
    3. Representation: according to the configuration, the representation for
       an element could be an integer (in this case, would be "Id"), or
       an one-hot vector (in this case, would be a list of integer).

    The class adopts the special elements from `Texar-Pytorch`, which are:

    1. <PAD>: which will be mapped into Id of 0 or -1 and have
       different representation according to different setting.
    2. <UNK>: if added into the vocabulary,
       will be the default element if the queried element is not found.

    Note that these two special tokens are necessary for the system in certain
    cases and thus must present in the vocabulary. The behavior of these
    special tokens are pre-defined based on different settings. To get around
    the default behavior (for example, if you have a pre-defined vocabulary
    with different setups), you can instruct the class to not adding these
    tokens automatically, and use the :func:`mark_special_element` instead.

    Here is a table on how our Vocabulary class behavior under different
    settings. Element0 means the first element that is added to the vocabulary.
    Elements added later will be element1, element2 and so on. They will follow
    the same behavior as element0. For readability, they are not
    listed in the table.

    .. list-table:: Vocabulary Behavior under different settings.

        * - `vocab_method`
          - custom (handle and implemented by the user)
          - indexing
          - indexing
          - one-hot
          - one-hot
        * - `need_pad`
          - assume False
          - True
          - False
          - True
          - False
        * - `get_pad_value`
          - None
          - 0
          - None
          - [0,0,0]
          - None
        * - `inner_mapping`
          - None
          - 0:pad 1:element0
          - 0:element0
          - -1:<PAD> 0:element0
          - 0:element0
        * - `element2repr`
          - raise Error
          - pad->0 element0->1
          - element0->0
          - <PAD>->[0,0,0] element0->[1,0,0]
          - element0->[1,0,0]
        * - `id2element`
          - raise Error
          - 0->pad 1->element0
          - 0->element0
          - -1 -> <PAD> 0->element0 (be careful)
          - 0->element0

    Args:
        method: The method to represent element in vocabulary, currently
            supporting "indexing" and "one-hot".
        use_pad: Whether to add <PAD> element to the vocabulary on
            creation. It will be added to the vocabulary first, but the id of
            it depends on the specific settings.
        use_unk: Whether to add <UNK> element to the vocabulary on
            creation. Elements that are not found in vocabulary will be
            directed to <UNK> element. It will be added right after the <PAD>
            element if provided.
        special_tokens: Additional special tokens to be added, they
            will be added at the beginning of vocabulary (but right after the
            <UNK> token) one by one.
        do_counting: Whether the vocabulary class will count the
            elements.
        pad_value: A customized value/representation to be used for
            padding, for example, following the PyTorch convention you may
            want to use -100. This value is only needed when `use_pad` is True.
            Default is None, where the value of padding is determined by
            the system.
        unk_value: A customized value/representation to be used for
            unknown value (`unk`). This value is only needed when `use_unk`
            is True. Default is None, where the value of `UNK` is determined
            by the system.

    Attributes:
        method (str): Same as above.
        use_pad (bool): Same as above.
        use_unk (bool): Same as above.
        do_counting (bool): Same as above.
    """

    def __init__(
        self,
        method: str = "indexing",
        use_pad: bool = True,
        use_unk: bool = True,
        special_tokens: Optional[List[str]] = None,
        do_counting: bool = True,
        pad_value: Any = None,
        unk_value: Any = None,
    ):
        self.method: str = method
        self.use_pad: bool = use_pad
        self.use_unk: bool = use_unk
        self.do_counting: bool = do_counting

        self._pad_id: Optional[int] = None
        self._unk_id: Optional[int] = None

        # Maps the raw element to the internal id.
        self._element2id: Dict = {}
        # Maps the internal id to the raw element.
        self._id2element: Dict = {}
        # Maps the internal id to the representation. This dict is populated
        #  when users provided customized representation of elements.
        self._id2repr: Dict = {}

        # Count the number of appearance of an element, indexed by the element
        # id.
        self.__counter: Counter = Counter()

        # Initialize the id auto counter.
        self.next_id = 0

        # Store the base special token names and their surface form.
        # By default, following the texar-pytorch special tokens:
        #   PAD: <PAD>
        #   UNK: <UNK>
        self._base_special_tokens: Dict[str, str] = {}

        # Store the id position of the special ids.
        self.__special_ids: Set[int] = set()

        if use_pad:
            # If not specified, will use -1 for padding in the case of
            #  one-hot. This will make the actual PAD representation to be
            #  a vector of zeros.
            pad_id = -1 if method == "one-hot" else None
            self.add_special_element(
                SpecialTokens.PAD,
                element_id=pad_id,
                special_token_name="PAD",
                representation=pad_value,
            )

        if use_unk:
            self.add_special_element(
                SpecialTokens.UNK,
                special_token_name="UNK",
                representation=unk_value,
            )

        if special_tokens is not None:
            for t in special_tokens:
                self.add_special_element(t)

    def get_count(self, e: Union[ElementType, int]) -> int:
        """
        Get the counts of the vocabulary element.

        Args:
            e: The element to get counts for. It can be the element id or the
              element's raw type.

        Returns:
            The count of the element.
        """
        if not self.do_counting:
            if not self.do_counting:
                raise InvalidOperationException(
                    "The vocabulary is not configured to count the elements."
                )

        eid: int = e if isinstance(e, int) else self._element2id[e]
        if self.is_special_token(eid):
            raise InvalidOperationException(
                "Count for special element is not available."
            )
        return self.__counter[eid]

    def mark_special_element(
        self, element_id: int, element_name: str, representation: Any = None
    ):
        """
        Mark a particular (but already existed) index in the vocabulary to be
        a special required element (i.e `PAD` or `UNK`).

        Args:
            element_id: The id to be set for the special element.
            element_name: The name of this element to be set, it can
              be one of `PAD`, `UNK`.
            representation: The representation/value that this element should
              be assigned. Default is None, then its representation will be
              computed from the internal indexing.
        """
        if element_name in ("PAD", "UNK"):
            if element_name == "PAD":
                self.use_pad = True
            if element_name == "UNK":
                self.use_unk = True
            if element_id in self._id2element:
                self._base_special_tokens[element_name] = self._id2element[
                    element_id
                ]
            else:
                raise ValueError(
                    f"Supplied {element_id} is not in the"
                    f" current vocabulary."
                )

            # Store the customized representation
            if representation is not None:
                self._id2repr[element_id] = representation
        else:
            raise ValueError(
                f"{element_name} is not a required special element, you can"
                f" add it in through `special_tokens` argument during class"
                f" creation, or calling the `add_special_element` method"
            )

    def is_special_token(self, element_id: int):
        """Check whether the element is a special token."""
        return element_id in self.__special_ids

    def add_special_element(
        self,
        element: str,
        element_id: Optional[int] = None,
        representation=None,
        special_token_name: Optional[str] = None,
    ):
        """
        This function will add special elements to the vocabulary, such as
        `UNK`, `PAD`, `BOS`, `CLS` symbols. Some special tokens will not be
        filtered by any `VocabFilter`. Some special tokens has their
        unique behavior in the system.

        .. note::

            most of the time, you don't have to call this method yourself,
            but should let the `init` function to handle that.

        Args:
            element: The surface form of this special element.
            element_id: The to be used for this special token.
                If not provided, the vocabulary will use the next id internally.
                If the provided id is occupied, a `ValueError` will be thrown.
                The id can be any integer, including negative ones.
            representation: The representation you want to assign to this
                special token. If None, the representation may be computed
                based on the index (which depends on the vocabulary setting).
            special_token_name: An internal name of
                this special token. This only matters for the base special
                tokens: <PAD> or <UNK>, and the name should be "PAD" and "UNK"
                respectively. Any other name here is considered invalid,
                and a `ValueError` will be thrown if provided.
        """
        if special_token_name is not None:
            if special_token_name not in ("PAD", "UNK"):
                raise ValueError(
                    "You don't have to and shouldn't provide the "
                    "`special_token_name` if this token is not PAD or UNK"
                )
            self._base_special_tokens[special_token_name] = element

        if element_id is not None:
            if element_id in self._id2element:
                raise ValueError(
                    f"ID {element_id} has already been used in Vocabulary. "
                )
        else:
            # Use auto-incremented id.
            element_id = self.__get_next_available_id()

        self._element2id[element] = element_id
        self._id2element[element_id] = element

        self.__special_ids.add(element_id)

        if representation is not None:
            self._id2repr[element_id] = representation

    def add_element(
        self, element: ElementType, representation: Any = None, count: int = 1
    ) -> int:
        r"""This function will add a regular element to the vocabulary.

        Args:
            element: The element to be added.
            representation: The vocabulary representation of this element
                will use this value. For example, you may want to use `-100`
                for ignored tokens for PyTorch skipped tokens. Note that the
                class do not check whether this representation is used by
                another element, so the caller have to manage the behavior
                itself.
            count: the count to be incremented for this element, default
                is 1 (i.e. consider it appear once on every add). This value
                will have effect only if `do_counting` is True.

        Returns:
            The internal id of the element.
        """
        element_id_: int
        try:
            element_id_ = self._element2id[element]
            if self.do_counting:
                self.__counter[element_id_] += count
        except KeyError:
            element_id_ = self.__get_next_available_id()
            self._element2id[element] = element_id_
            self._id2element[element_id_] = element
            if representation:
                self._id2repr[element_id_] = representation
            if self.do_counting:
                self.__counter[element_id_] = count

        return element_id_

    def __get_next_available_id(self):
        """Find the next available id by incrementing the auto counter until
        one is found.
        """
        eid = self.next_id
        while eid in self._id2element:
            self.next_id += 1
            eid = self.next_id
        return eid

    def id2element(self, idx: int) -> ElementType:
        r"""This function will map id to element.

        Args:
            idx: The queried id of element.

        Returns:
            The corresponding element if exist. Check the behavior
            of this function under different setting in the documentation.

        Raises:
            KeyError: If the id is not found.
        """
        return self._id2element[idx]

    def element2repr(
        self, element: Union[ElementType, Any]
    ) -> Union[int, List[int]]:
        r"""This function will map element to representation.

        Args:
            element: The queried element. It can be either the same
              type as the element, or string (for the special tokens).

        Returns:
            Union[int, List[int]]: The corresponding representation
            of the element. Check the behavior of this function
            under different setting in the documentation.

        Raises:
            KeyError: If element is not found and vocabulary does
                not use <UNK> element.
        """
        if self.use_unk:
            idx = self._element2id.get(
                element, self._element2id[self._base_special_tokens["UNK"]]
            )
        else:
            idx = self._element2id[element]

        # If a custom representation is set for this idx, we will use it.
        if idx in self._id2repr:
            return self._id2repr[idx]
        elif self.method == "indexing":
            return idx
        elif self.method == "one-hot":
            return self._one_hot(idx)
        else:
            raise InvalidOperationException(
                f"Cannot find the representation for idx at [{idx}], it does"
                f" not have a customized representation, and the representation"
                f" method [{self.method}] is not supported."
            )

    def to_dict(self) -> Dict[ElementType, Any]:
        """
        Create a dictionary from the vocabulary storing all the known elements.

        Returns:
            The vocabulary as a Dict from ElementType to the
            representation of the element (could be Integer or One-hot vector,
            depending on the settings of this class).
        """
        vocab_dict: Dict[ElementType, Any] = {}
        for element in self._element2id:
            vocab_dict[element] = self.element2repr(element)
        return vocab_dict

    def _one_hot(self, idx: int):
        """Compute the one-hot encoding on the fly."""
        vec_size = len(self._element2id)
        if self.use_pad:
            vec_size -= 1
        vec = [0 for _ in range(vec_size)]
        if idx != -1:
            vec[idx] = 1
        return vec

    def __len__(self) -> int:
        r"""This function return the size of vocabulary.

        Returns:
            int: The number of elements, including <PAD>, <UNK>.
        """
        return len(self._element2id)

    def has_element(self, element: Union[ElementType, str]) -> bool:
        r"""This function checks whether an element is added to vocabulary.

        Args:
            element: The queried element.

        Returns:
            bool: Whether element is found.
        """
        return element in self._element2id

    def vocab_items(self) -> Iterable[Tuple[Union[ElementType, str], int]]:
        r"""This function will loop over the (element, id) pair inside this
        class.

        Returns:
            Iterable[Tuple]: Iterables of (element, id) pair.
        """
        return self._element2id.items()

    def get_pad_value(self) -> Union[None, int, List[int]]:
        r"""This function will get the representation of the PAD element for
        the vocabulary. The representation depends on the settings of this
        class, it can be an integer or a list of int (e.g. a vector).

        Returns:
            Union[None, int, List[int]]: The PAD element. Check
            the behavior of this function in the class documentation.
        """
        if self.use_pad:
            return self.element2repr(self._base_special_tokens["PAD"])
        return None

    def filter(self, vocab_filter: "VocabFilter") -> "Vocabulary":
        """
        This function will create a new vocabulary object, which
        is based on the current vocabulary, but filter out elements that
        appear fewer times than the `min_count` value. Calling this function
        will cause a full iteration over the vocabulary, thus normally, it
        should be called after collecting all the vocabulary in the dataset.

        Args:
            vocab_filter: The filter used to filter the vocabulary.

        Returns:
            A new vocabulary after filtering.
        """
        # Make a new vocab class:
        # 1. We do not add the PAD or UNK at init, but will copy them later.
        # 2. We also ignore other special tokens at init, but copy them later.
        # 2. We follow the do_counting setup.
        vocab: Vocabulary = Vocabulary(
            self.method,
            use_pad=False,
            use_unk=False,
            do_counting=self.do_counting,
        )
        # We then set these flag manually based on this vocabulary.
        vocab.use_pad = self.use_pad
        vocab.use_unk = self.use_unk

        # Now we copy all the vocabulary items to the new vocab.
        for element, eid in self.vocab_items():
            # Copy the special tokens regardless of the filter.
            if self.is_special_token(eid):
                element_name = None
                try:
                    if element == self._base_special_tokens["PAD"]:
                        element_name = "PAD"
                except KeyError:
                    # No PAD in the origin vocab.
                    pass

                try:
                    if element == self._base_special_tokens["UNK"]:
                        element_name = "UNK"
                except KeyError:
                    # No UNK in the origin vocab.
                    pass

                # Special element value must be string.
                assert isinstance(element, str)
                vocab.add_special_element(
                    element, eid, self._id2repr.get(element, None), element_name
                )
            elif not vocab_filter.filter(eid):
                vocab.add_element(
                    element,
                    count=self.get_count(eid) if self.do_counting else 1,
                )
        return vocab


class VocabFilter(ABC):
    """
    Base class for vocabulary filters, which is used to implement constraints
    to choose a subset of vocabulary. For example, one can filter out vocab
    elements that happen fewer than a certain frequency.

    Args:
        vocab: The vocabulary object to be filtered.
    """

    def __init__(self, vocab: Vocabulary):
        self._vocab = vocab

    def filter(self, element_id: int) -> bool:
        """
        Given the element id, it will determine whether the element should be
        filtered out.

        Args:
            element_id: The element id to be checked.

        Returns:
            None
        """
        raise NotImplementedError


class FrequencyVocabFilter(VocabFilter):
    """
    A frequency based filter. It will filter vocabulary elements that appear
    fewer than `min_frequency` or more than `max_frequency`. The check will
    be skipped if the threshold values are negative.

    Args:
        vocab: The vocabulary object.
        min_frequency: The min frequency threshold, default -1 (i.e. no
          frequency check for min).
        max_frequency: The max frequency threshold, default -1 (i.e. no
          frequency check for max).

    """

    def __init__(
        self,
        vocab: Vocabulary,
        min_frequency: int = -1,
        max_frequency: int = -1,
    ):
        super().__init__(vocab)
        self.min_freq = min_frequency
        self.max_freq = max_frequency

        if not vocab.do_counting:
            raise InvalidOperationException(
                "The provided vocabulary is not configured to collect counts, "
                "cannot filter the vocabulary based on counts."
            )

    def filter(self, element_id: int) -> bool:
        freq = self._vocab.get_count(element_id)

        will_filter = False
        if self.min_freq >= 0 and freq < self.min_freq:
            will_filter = True

        if 0 <= self.max_freq < freq:
            will_filter = True

        return will_filter



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from forte.utils import create_import_error_msg

__all__ = ["Dictionary", "WordnetDictionary"]


class Dictionary:
    r"""
    This class defines a dictionary for word replacement.
    Given an input word and its pos_tag(optional), the dictionary
    will outputs its synonyms, antonyms, hypernyms and hypernyms.
    """

    # pylint: disable=unused-argument
    def get_synonyms(
        self, word: str, pos_tag: str = "", lang: str = "eng"
    ) -> List[str]:
        r"""
        Args:
            word: The input string.
            pos_tag: The Part-of-Speech tag for substitution.
            lang: The language of the input string.
        Returns:
            synonyms of the word.
        """
        return []

    def get_antonyms(
        self, word: str, pos_tag: str = "", lang: str = "eng"
    ) -> List[str]:
        r"""
        Args:
            word: The input string.
            pos_tag: The Part-of-Speech tag for substitution.
            lang: The language of the input string.
        Returns:
            Antonyms of the word.
        """
        return []

    def get_hypernyms(
        self, word: str, pos_tag: str = "", lang: str = "eng"
    ) -> List[str]:
        r"""
        Args:
            word: The input string.
            pos_tag: The Part-of-Speech tag for substitution.
            lang: The language of the input string.
        Returns:
            Hypernyms of the word.
        """
        return []

    def get_hyponyms(
        self, word: str, pos_tag: str = "", lang: str = "eng"
    ) -> List[str]:
        r"""
        Args:
            word: The input string.
            pos_tag: The Part-of-Speech tag for substitution.
            lang: The language of the input string.
        Returns:
            Hyponyms of the word.
        """
        return []


class WordnetDictionary(Dictionary):
    r"""
    This class wraps the nltk WORDNET to replace
    the input word with an synonym/antonym/hypernym/hyponym.
    Part-of-Speech(optional) can be provided to the wordnet
    for retrieving words with the same POS.
    """

    def __init__(self):
        try:
            import nltk  # pylint: disable=import-outside-toplevel
            from nltk.corpus import (  # pylint:disable=import-outside-toplevel
                wordnet,
            )
        except ImportError as err:
            raise ImportError(
                create_import_error_msg(
                    "nltk", "data_aug", "dictionary based data augmentation"
                )
            ) from err

        try:
            # Check if the wordnet package and
            # pos_tag package are downloaded.
            wordnet.synsets("computer")
        except LookupError:
            nltk.download("wordnet")
            nltk.download("omw-1.4")
        self.model = wordnet

    def _get_wordnet_pos(self, treebank_tag: str) -> str:
        """
        return WORDNET POS compliance to WORDNET lemmatization (a,n,r,v)
        """
        if treebank_tag.startswith("J"):
            return self.model.ADJ
        elif treebank_tag.startswith("V"):
            return self.model.VERB
        elif treebank_tag.startswith("N"):
            return self.model.NOUN
        elif treebank_tag.startswith("R"):
            return self.model.ADV
        else:
            # As default pos in lemmatization is Noun
            return self.model.NOUN

    def get_lemmas(
        self,
        word: str,
        pos_tag: str = "",
        lang: str = "eng",
        lemma_type: str = "SYNONYM",
    ):
        r"""
        This function gets synonyms/antonyms/hypernyms/hyponyms
        from a WORDNET dictionary.

        Args:
            word: The input token.
            pos_tag: The NLTK POS tag.
            lang: The input language.
            lemma_type: The type of words to replace, must be
                one of the following:

                - ``'SYNONYM'``
                - ``'ANTONYM'``
                - ``'HYPERNYM'``
                - ``'HYPONYM'``
        """
        res: List[str] = []
        pos_wordnet = None
        # The POS property is used for retrieving lemmas with the same POS.
        if pos_tag and len(pos_tag) > 0:
            pos_wordnet = self._get_wordnet_pos(pos_tag)

        for synonym in self.model.synsets(word, pos=pos_wordnet, lang=lang):
            for lemma in synonym.lemmas(lang=lang):
                if lemma_type == "SYNONYM":
                    res.append(lemma.name())
                elif lemma_type == "ANTONYM":
                    for antonym in lemma.antonyms():
                        res.append(antonym.name())
                elif lemma_type == "HYPERNYM":
                    for hypernym in lemma.hypernyms():
                        res.append(hypernym.name())
                elif lemma_type == "HYPONYM":
                    for hyponym in lemma.hyponyms():
                        res.append(hyponym.name())
                else:
                    raise KeyError(
                        "The type {} does not belong to "
                        '["SYNONYM", "ANTONYM", '
                        '"HYPERNYM", "HYPONYM"]]'.format(type)
                    )
        # The phrases are concatenated with "_" in wordnet.
        return [word.replace("_", " ") for word in res]

    def get_synonyms(
        self, word: str, pos_tag: str = "", lang: str = "eng"
    ) -> List[str]:
        r"""
        This function replaces a word with synonyms from a WORDNET dictionary.
        """
        return self.get_lemmas(word, pos_tag, lang, lemma_type="SYNONYM")

    def get_antonyms(
        self, word: str, pos_tag: str = "", lang: str = "eng"
    ) -> List[str]:
        r"""
        This function replaces a word with antonyms from a WORDNET dictionary.
        """
        return self.get_lemmas(word, pos_tag, lang, lemma_type="ANTONYM")

    def get_hypernyms(
        self, word: str, pos_tag: str = "", lang: str = "eng"
    ) -> List[str]:
        r"""
        This function replaces a word with hypernyms from a WORDNET dictionary.
        """
        return self.get_lemmas(word, pos_tag, lang, lemma_type="HYPERNYM")

    def get_hyponyms(
        self, word: str, pos_tag: str = "", lang: str = "eng"
    ) -> List[str]:
        r"""
        This function replaces a word with hyponyms from a WORDNET dictionary.
        """
        return self.get_lemmas(word, pos_tag, lang, lemma_type="HYPONYM")



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base class for processors.
"""
import logging
from abc import abstractmethod, ABC
from typing import Any, Dict, Set

from forte.data.base_pack import PackType
from forte.data.selector import DummySelector
from forte.pipeline_component import PipelineComponent
from forte.utils.utils_processor import (
    record_types_and_attributes_check,
    collect_input_pack_record,
)

__all__ = [
    "BaseProcessor",
]


class BaseProcessor(PipelineComponent[PackType], ABC):
    r"""Base class inherited by all kinds of processors such as trainer,
    predictor and evaluator.
    """

    def __init__(self):
        super().__init__()
        self.selector = DummySelector()

    def record(self, record_meta: Dict[str, Set[str]]):
        r"""Method to add output record of the current processor to
        :attr:`forte.data.data_pack.Meta.record`. The key of the record
        should be the entry type and values should be attributes of the entry
        type. All the information would be used for consistency checking
        purpose if the pipeline is initialized with
        `enforce_consistency=True`.

        Args:
            record_meta: The field in the datapack for type record that need to
                fill in for consistency checking.
        """
        pass

    def expected_types_and_attributes(self) -> Dict[str, Set[str]]:
        r"""Method to add expected types and attributes for the input of the
        current processor which would be checked before running the processor if
        if the pipeline is initialized with
        `enforce_consistency=True`.
        """

        return {}

    def check_record(self, input_pack: PackType):
        # pylint: disable=protected-access
        r"""Method to check type consistency if the pipeline is initialized with
        `enforce_consistency=True`. If any expected type or its attribute
        does not exist in the datapack record of the previous pipeline
        component, an error of
        :class:`~forte.common.exception.ExpectedRecordNotFound` will be raised.

        Args:
            input_pack: The input datapack.
        """
        if self._check_type_consistency:
            expectation = self.expected_types_and_attributes()
            input_pack_record = collect_input_pack_record(
                self.resources, input_pack
            )
            record_types_and_attributes_check(expectation, input_pack_record)

    def write_record(self, input_pack: PackType):
        r"""Method to write records of the output type of the current
        processor to the datapack. The key of the record should be the entry
        type and values should be attributes of the entry type. All the
        information would be used for consistency checking purpose if
        the pipeline is initialized with
        `enforce_consistency=True`.

        Args:
            input_pack: The input datapack.

        """
        # pylint: disable=protected-access
        if self._check_type_consistency:
            try:
                self.record(input_pack._meta.record)
            except AttributeError:
                # For backward compatibility, no record to write.
                logging.info(
                    "Packs of the old format do not have the record field."
                )

    def process(self, input_pack: PackType):
        self.check_record(input_pack)
        # Set the component for recording purpose.
        self._process(input_pack)
        self.write_record(input_pack)

    @abstractmethod
    def _process(self, input_pack: PackType):
        r"""The main function of the processor. The implementation should
        process the ``input_pack``, and conduct operations such as adding
        entries into the pack, or produce some side-effect such as writing
        data into the disk.

        Args:
            input_pack: The input datapack.
        """
        raise NotImplementedError

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        r"""Returns a `dict` of configurations of the processor with default
        values. Used to replace the missing values of input ``configs`` during
        pipeline construction.
        """
        return {
            "overwrite": False,
        }



#  Copyright 2021 The Forte Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pickle
from typing import Dict, Any, Optional

from forte.common.configuration import Config
from forte.data import BaseExtractor
from forte.data.converter import Converter
from forte.utils import get_class

DATA_INPUT = 0
DATA_OUTPUT = 1


def parse_feature_extractors(scheme_configs: Config) -> Dict[str, Any]:
    feature_requests: Dict[str, Any] = {}

    for tag, scheme_config in scheme_configs.items():
        assert (
            "extractor" in scheme_config
        ), "Field not found for data request scheme: `extractor`"
        assert (
            "type" in scheme_config
        ), "Field not found for data request scheme: `type`"
        assert scheme_config["type"] in [
            "data_input",
            "data_output",
        ], "Type field must be either data_input or data_output."

        feature_requests[tag] = {}

        if scheme_config["type"] == "data_input":
            feature_requests[tag]["type"] = DATA_INPUT
        elif scheme_config["type"] == "data_output":
            feature_requests[tag]["type"] = DATA_OUTPUT

        extractor_class = get_class(scheme_config["extractor"]["class_name"])
        extractor: BaseExtractor = extractor_class()
        if not isinstance(extractor, BaseExtractor):
            raise RuntimeError(
                "Invalid extractor: ", scheme_config["extractor"]
            )

        extractor.initialize(config=scheme_config["extractor"]["config"])

        # Load vocab from disk if provided.
        if "vocab_path" in scheme_config["extractor"]:
            with open(
                scheme_config["extractor"]["vocab_path"], "rb"
            ) as vocab_file:
                extractor.vocab = pickle.load(vocab_file)

        feature_requests[tag]["extractor"] = extractor

        if "converter" not in scheme_config:
            # Create default converter if there is no given converter
            feature_requests[tag]["converter"] = Converter({})
        else:
            converter_class = get_class(
                scheme_config["converter"]["class_name"]
            )
            converter: Converter = converter_class()
            if not isinstance(converter, Converter):
                raise RuntimeError(
                    "Invalid converter: ", scheme_config["converter"]
                )
            feature_requests[tag]["converter"] = converter

    return feature_requests


def add_extractor(
    request: Dict,
    name: str,
    extractor: BaseExtractor,
    is_input: bool,
    converter: Optional[Converter] = None,
):
    """
    Extractors can be added to the preprocessor directly via this
    method.

    Args:
        request: A request dictionary to be populated.
        name: The name/identifier of this extractor, the name should be
          different between different extractors.
        extractor: The extractor instance to be added.
        is_input: Whether this extractor will be used as input or output.
        converter:  The converter instance to be applied after running
          the extractor.

    Returns:

    """
    request["schemes"][name]["extractor"] = extractor
    request["schemes"][name]["type"] = DATA_INPUT if is_input else DATA_OUTPUT
    request["schemes"][name]["converter"] = (
        Converter({}) if converter is None else converter
    )



# Copyright 2021 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=attribute-defined-outside-init
"""
The module contains assorted common readers.
"""
import logging
from abc import ABC
from typing import Iterator, Dict, Any

from forte.data import DataPack, MultiPack
from forte.data.base_pack import PackType
from forte.data.base_reader import PackReader, BaseReader
from ft.onto.base_ontology import Utterance

logger = logging.getLogger(__name__)

__all__ = [
    "TerminalReader",
    "RawPackReader",
    "RawMultiPackReader",
]


class TerminalReader(PackReader):
    r"""A reader designed to read text from the terminal."""

    # pylint: disable=unused-argument
    def _cache_key_function(self, collection) -> str:
        return "cached_string_file"

    def _collect(self) -> Iterator[str]:  # type: ignore
        # This allows the user to pass in either one single string or a list of
        # strings.
        while True:
            try:
                data = input(self.configs.prompt_text)
                if len(data) == 0:
                    continue
                yield data
            except EOFError:
                print()
                break

    def _parse_pack(self, data_source: str) -> Iterator[DataPack]:
        pack = DataPack()
        self.set_text(pack, data_source)
        Utterance(pack, 0, len(data_source))
        yield pack

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        """
        Defines the default configuration for this class, available options are:

        .. code-block:: python

            {
                "prompt_text": "Enter your query here: ",
            }

        Here:

        `prompt_text` defines the text shown on the terminal as a prompt
        for the user.

        Returns: The default configuration values as a dict.
        """
        configs = super().default_configs()
        configs.update(
            {
                "pack_name": "query",
                "prompt_text": "Enter your query here: ",
            }
        )

        configs["pack_name"] = "query"
        return configs


class BaseRawPackReader(BaseReader, ABC):
    def _collect(self, pack: PackType) -> Iterator[Any]:  # type: ignore
        yield pack

    def _parse_pack(self, pack: PackType) -> Iterator[PackType]:
        yield pack


class RawPackReader(BaseRawPackReader):
    @staticmethod
    def pack_type():
        return DataPack


class RawMultiPackReader(BaseRawPackReader):
    @staticmethod
    def pack_type():
        return MultiPack



# Copyright 2022 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import csv
import importlib
from typing import Iterator, Tuple, List, Dict
from forte.common import Resources, ProcessorConfigError
from forte.common.configuration import Config
from forte.data.data_pack import DataPack
from forte.data.base_reader import PackReader
from ft.onto.base_ontology import *

__all__ = ["ClassificationDatasetReader"]


class ClassificationDatasetReader(PackReader):
    r""":class:`ClassificationDatasetReader` is designed to read table-like
    classification datasets that contain input text and digit/text labels.
    There are a couple of values that need to be provided via configuration,
    including `forte_data_fields`, `text_fields`, and `index2class`.

    `forte_data_fields` is a list representing the column headers of the
    dataset. Each element in the `forte_data_fields` list is either a label
    that indicates the body class or a Forte entry type that will be used
    to store the content. Apparently, the list should follow the column order
    of the dataset.

    For example, for amazon polarity dataset,
    (https://huggingface.co/datasets/amazon_polarity), the column names
    are [label, title, content]. We can configure `forte_data_fields`to be
    ['label', 'ft.onto.base_ontology.Title', 'ft.onto.base_ontology.Body'].
    `label` is a special keyword to specify the label/body class column,
    while the latter two ontology types will be used by the reader to store the
    text from the `title` and `content` column respectively.

    `text_fields` is a list of Forte entry types that indicate texts in the
    forte data fields will be kept and concatenated as input text.
    Apparently, it's a subset of `forte_data_fields`.
    For example, if we only want titles and bodies in our input text, we
    specify "text_fields":
    ['ft.onto.base_ontology.Title', 'ft.onto.base_ontology.Body']" in the
    configuration. If titles are not needed, we can also only include
    bodies by specifying
    "text_fields":['ft.onto.base_ontology.Body']" which is very
    flexible in customizing input text.


    `index2class` is a dictionary that the mapping from zero-based
    indices to classes. For example, in amazon polarity dataset, we have two
    classes, negative and positive.
    We can configure `index2class` to be `{0: negative, 1: positive}`.
    sentiment classifications.

    To see a full example, please refer to
    https://github.com/asyml/forte/examples/classification
    """

    def __init__(self):
        super().__init__()
        self._index2class = None
        self._class2index = None

    def set_up(self):
        if self.configs.index2class is None:
            raise ProcessorConfigError(
                "User must set index2class to enable"
                " the dataset reader to encode labels correctly."
                "Current index2class is set to "
                f"{self.configs.index2class}"
            )
        if self.configs.skip_k_starting_lines < 0:
            raise ProcessorConfigError(
                "Current skip_k_starting_lines in configuration is set to "
                f"{self.configs.skip_k_starting_lines}. User must specify an"
                " integer larger or equal to 0."
            )
        # class and index
        if not self.configs.digit_label:
            # initialize class
            self._index2class = self.configs.index2class
            self._class2index = {v: k for k, v in self._index2class.items()}

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)
        self.set_up()
        if "label" not in self.configs.forte_data_fields:
            raise ProcessorConfigError(
                "There must be a forte data field named 'label'"
                " in the reader config."
            )

        if "ft.onto.base_ontology.Document" in self.configs.forte_data_fields:
            raise ProcessorConfigError(
                "Found `ft.onto.base_ontology.Document` in forte_data_fields "
                "of the reader config. Try use a different ontology to "
                "represent the data field as this ontology is reserved to "
                "store the concatenated text from input."
            )

        if self.configs.text_fields is None:
            self.configs.text_fields = [
                df
                for df in self.configs.forte_data_fields
                if df is not None and df != "label"
            ]
        else:
            if not set(self.configs.text_fields).issubset(
                set(self.configs.forte_data_fields)
            ):
                raise ProcessorConfigError(
                    "text_fields must be a subset of forte_data_fields."
                    f"text_fields: {self.configs.text_fields}"
                    f"forte_data_fields: {self.configs.forte_data_fields}"
                    "Please correct text_fields and forte_data_fields in the"
                    " configuration to satisfy the condition."
                )

    def _collect(  # type: ignore
        self, csv_file: str
    ) -> Iterator[Tuple[int, List[str]]]:
        with open(csv_file, encoding="utf-8") as f:
            data = csv.reader(f, delimiter=",", quoting=csv.QUOTE_ALL)
            if self.configs.skip_k_starting_lines > 0:
                for _ in range(self.configs.skip_k_starting_lines):
                    next(data)
            for line_id, line in enumerate(data):
                yield line_id, line

    def _cache_key_function(self, line_info: Tuple[int, List[str]]) -> str:
        return str(line_info[0])

    def _parse_pack(
        self, line_info: Tuple[int, List[str]]
    ) -> Iterator[DataPack]:
        line_id, line = line_info
        # content label title
        pack = DataPack()

        if len(self.configs.forte_data_fields) != len(line):
            raise ProcessorConfigError(
                "Forte data fields provided in config "
                "doesn't have one-to-one correspondence with the actual data "
                "fields from dataset.\n"
                f"Data fields length: {len(self.configs.forte_data_fields)} \n"
                f"Line length: {len(line)}"
            )
        df_dict = dict(zip(self.configs.forte_data_fields, line))

        # it determines the order of concatenation
        text_fields = self.configs.text_fields

        # get text and ontology indices
        text, input_ontology_indices = generate_text_n_input_ontology_indices(
            text_fields, df_dict
        )
        self.set_text(pack, text)
        if df_dict["label"].isdigit() != self.configs.digit_label:
            dataset_digit_label = df_dict["label"].isdigit()
            raise ProcessorConfigError(
                "Label format from dataset "
                "is not consistent with the label format from configs. \n"
                f"dataset digit label: {dataset_digit_label} \n"
                f"config digit label: {self.configs.digit_label}",
            )

        if self.configs.digit_label:
            if self.configs.one_based_index_label:
                # adjust label encoding if dataset digit label is one-based
                class_id = int(df_dict["label"]) - 1
            else:
                class_id = int(df_dict["label"])
        else:
            # decode string label from dataset into digit label
            class_id = self._class2index[df_dict["label"]]
        # initialize all ontologies instances as data pack module
        # to add corresponding variables and functions
        for (
            ontology_field,
            (start_idx, end_idx),
        ) in input_ontology_indices.items():
            path_str, module_str = ontology_field.rsplit(".", 1)
            mod = importlib.import_module(path_str)  # sentence ontology module
            entry_class = getattr(mod, module_str)
            entry_class(pack, start_idx, end_idx)
        # for now, we use Document to store concatenated text and set the class here
        doc = Document(pack, 0, input_ontology_indices[text_fields[-1]][1])
        doc.document_class = [self.configs.index2class[class_id]]

        pack.pack_name = line_id
        yield pack

    @classmethod
    def default_configs(cls):
        r"""This defines a basic configuration structure for classification dataset reader.

        Here:

          - forte_data_fields: these fields provides one-to-one
            correspondence between given original dataset column names and
            labels or forte ontology. For column names without usage,
            user can specify None for them.
          - index2class: a dictionary that maps from zero-based indices to
            classes
          - text_fields: a list of ordered input ontology that
            user want to concatenate into an input text.
          - digit_label:  boolean value that specifies whether label in dataset
            is digit.
          - one_based_index_label: boolean value that specifies if dataset
            provides one-based digit label.
            True for one-based index, false otherwise.
          - skip_k_starting_lines: many datasets' first line are columns
            names, set it to 1 if it's the case. Otherwise set to 0.
            User can also set it to other positive integers to skip
            multiple lines.
        """
        return {
            "forte_data_fields": [
                "label",
                "ft.onto.base_ontology.Title",
                "ft.onto.base_ontology.Body",
            ],
            "index2class": None,
            "text_fields": None,
            "digit_label": True,
            "one_based_index_label": True,
            "skip_k_starting_lines": 1,
        }


def generate_text_n_input_ontology_indices(
    text_fields: List[str], forte_data_fields_dict: Dict[str, str]
) -> Tuple[str, Dict[str, Tuple[int, int]]]:
    """
    Retrieve ontologies from data fields and concatenate them into text.
    Also, we generate the indices for these ontologies accordingly.

    Args:
        text_fields(List[str]): a list of ontology that needs to be
            concatenated into a input string.
        forte_data_fields_dict(Dict[str, str]): a dictionary with ontology names as keys and
            ontology strings as values.

    Returns:
        Tuple[str, Dict[str, Tuple[int, int]]]: a concatenated text and
            dictionary that keys are forte data entries and values are start
            and end indices of the data entries.
    """
    end = -1
    text = ""
    indices = {}  # a dictionary of (ontology_name: (start_index, end_index) )
    for i, input_onto in enumerate(text_fields):
        text += forte_data_fields_dict[input_onto]
        start = end + 1
        end = len(text)
        if i != len(text_fields) - 1:  # not last input ontology
            text += "\n"
        indices[text_fields[i]] = (start, end)

    return text, indices



# Copyright 2022 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This Processor focuses on calling data augmentation ops
to generate texts similar to those in the input pack
and create a new pack with them.
"""
from collections import defaultdict
from typing import List, Tuple, Dict, DefaultDict, Union, cast
from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack
from forte.data.ontology.core import Entry, BaseLink
from forte.data.ontology.top import (
    MultiPackLink,
    MultiPackGroup,
)
from forte.processors.base import MultiPackProcessor
from forte.utils.utils import create_class_with_kwargs
from forte.processors.data_augment.algorithms.base_data_augmentation_op import (
    BaseDataAugmentationOp,
)

__all__ = ["BaseOpProcessor"]


class BaseOpProcessor(MultiPackProcessor):
    r"""
    This is a Base Data Augmentation Op Processor that instantiates
    data augmentation ops into Forte Data Structures to be used. It can
    handle augmentations of multiple ontology types simultaneously and
    copy other existing Forte entries based on policies specified in
    `other_entry_policy` configuration from source data pack to
    augmented data pack.
    """

    def __init__(self):
        super().__init__()

        # :attr:`_new_data_packs`: {datapack id: datapack}
        # It records the mapping from the original data pack
        # to the augmented data pack
        self._new_data_packs: DefaultDict[int, DataPack] = defaultdict()

        # :attr:`_data_pack_map`: {orig pack id: new pack id}
        # It maintains a mapping from the pack id
        # of the original pack to the pack id of augmented pack.
        # It is used when copying the MultiPackLink and MultiPackGroup.
        self._data_pack_map: Dict[int, int] = {}

        # :attr:`_entry_maps`: {datapack id: Dict{orig tid, new tid}}
        # It is a map for tracking the annotation ids
        # before and after the auto align. It maps the
        # original annotation tid to the new annotation tid.
        # It is used when copying the Link/Group/MultiPackLink/MultiPackGroup.
        self._entry_maps: Dict[int, Dict[int, int]] = {}

        # :attr:`replacement_op`: BaseAugmentationOp
        # It is the augmentation Op used by this processor
        # instance.
        self.replacement_op: BaseDataAugmentationOp = None

        self.configs = None

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)
        self.configs = self.make_configs(configs)

    def _copy_multi_pack_link_or_group(
        self, entry: Union[MultiPackLink, MultiPackGroup], multi_pack: MultiPack
    ) -> bool:
        r"""
        This function copies a MultiPackLink/MultiPackGroup in the multipack.
        It could be used in tasks such as text generation, where
        MultiPackLink is used to align the source and target.

        Args:
            entry: The MultiPackLink/MultiPackGroup to copy.
            multi_pack: The multi_pack contains the input entry.

        Returns:
            A bool value indicating whether the copy happens.
        """
        # The entry should be either MultiPackLink or MultiPackGroup.
        is_link: bool = isinstance(entry, BaseLink)
        children: List[Entry]
        if is_link:
            children = [entry.get_parent(), entry.get_child()]  # type: ignore
        else:
            children = entry.get_members()  # type: ignore

        # Get the copied children entries.
        new_children: List[Entry] = []
        for child_entry in children:
            child_pack: DataPack = child_entry.pack
            child_pack_pid: int = child_pack.pack_id
            # The new pack should be present.
            if (
                child_pack_pid not in self._data_pack_map
                or child_pack_pid not in self._entry_maps
            ):
                return False
            new_child_pack: DataPack = multi_pack.get_pack_at(
                multi_pack.get_pack_index(self._data_pack_map[child_pack_pid])
            )
            # The new child entry should be present.
            if child_entry.tid not in self._entry_maps[child_pack_pid]:
                return False
            new_child_tid: int = self._entry_maps[child_pack_pid][
                child_entry.tid
            ]
            new_child_entry: Entry = new_child_pack.get_entry(new_child_tid)
            new_children.append(new_child_entry)

        # Create the new entry and add to the multi pack.
        new_entry: Entry
        if is_link:
            entry = cast(MultiPackLink, entry)

            new_link_parent, new_link_child = new_children

            new_entry = type(entry)(
                multi_pack, new_link_parent, new_link_child  # type: ignore
            )
        else:
            entry = cast(MultiPackGroup, entry)
            new_entry = type(entry)(multi_pack, new_children)  # type: ignore
        multi_pack.add_entry(new_entry)
        return True

    def _clear_states(self):
        r"""
        This function clears the states. It should be
        called after processing a multipack.
        """
        self.replacement_op.clear_states()
        self._data_pack_map.clear()
        self._entry_maps.clear()

    def _augment(
        self, input_pack: MultiPack, aug_pack_names: List[str]
    ) -> bool:
        r"""
        This function calls the data augmentation ops and
        modifies the input in-place. The subclasses should override
        this function to implement other data augmentation methods, such
        as Easy Data Augmentation.

        Args:
            input_pack: The input MultiPack.
            aug_pack_names: The packs names for DataPacks to be augmented.

        Returns:
            A boolean value indicating if the data augmentation was
            sucessful or not.
        """
        try:
            self.replacement_op = create_class_with_kwargs(
                self.configs["data_aug_op"],
                class_args={"configs": self.configs["data_aug_op_config"]},
            )

            for pack_name in aug_pack_names:
                data_pack: DataPack = input_pack.get_pack(pack_name)
                self._new_data_packs[
                    data_pack.pack_id
                ] = self.replacement_op.perform_augmentation(data_pack)

            (
                self._data_pack_map,
                self._entry_maps,
            ) = self.replacement_op.get_maps()
            return True
        except ValueError:
            return False

    def _process(self, input_pack: MultiPack):
        # Get the pack names for augmentation.
        aug_pack_names: List[str] = []

        # Check if the DataPack exists.
        for pack_name in self.configs["augment_pack_names"].keys():
            if pack_name in input_pack.pack_names:
                aug_pack_names.append(pack_name)

        if len(self.configs["augment_pack_names"].keys()) == 0:
            # Augment all the DataPacks if not specified.
            aug_pack_names = list(input_pack.pack_names)

        success = self._augment(input_pack, aug_pack_names)

        if not success:
            raise ValueError(
                "There was a problem encountered when performing the augmentation in {}".format(
                    self.configs["data_aug_op"]
                )
            )

        new_packs: List[Tuple[str, DataPack]] = []

        for aug_pack_name in aug_pack_names:
            new_pack_name: str = self.configs["augment_pack_names"].get(
                aug_pack_name, "augmented_" + aug_pack_name
            )
            data_pack = input_pack.get_pack(aug_pack_name)

            new_packs.append(
                (new_pack_name, self._new_data_packs[data_pack.pack_id])
            )

        for new_pack_name, new_pack in new_packs:
            input_pack.add_pack_(new_pack, new_pack_name)

        # Copy the MultiPackLinks/MultiPackGroups
        for mpl in input_pack.get(MultiPackLink):
            self._copy_multi_pack_link_or_group(mpl, input_pack)
        for mpg in input_pack.get(MultiPackGroup):
            self._copy_multi_pack_link_or_group(mpg, input_pack)

        # Must be called after processing each multipack
        # to reset internal states.
        self._clear_states()

    @classmethod
    def default_configs(cls):
        """
        Returns:
            A dictionary with the default config for this processor.
        Following are the keys for this dictionary:
            - data_aug_op:
                The data augmentation Op for the processor.
                It should be a full qualified class name.
                Example:
                    "forte.processors.data_augment.algorithms.
                    text_replacement_op.TextReplacementOp"
            - data_aug_op_config:
                The configuration for data augmentation Op.
                Example:
                    .. code-block:: python
                        'data_aug_op_config': {
                            'lang': 'en',
                            'use_gpu': False,
                            'other_entry_policy': {
                                'ft.onto.base_ontology.Document': 'auto_align',
                                'ft.onto.base_ontology.Sentence': 'auto_align',
                            }
                        }
            - augment_pack_names:
                A dict specifies the DataPacks to augment and their output
                names. It should be key-value pairs where the key is the
                input DataPack name, and the value is the output DataPack
                name after augmentation.
                If empty, all the DataPacks will be augmented, and the output
                names will be automatically generated by prepending
                an `'augmented_'` prefix.
                Example:
                    .. code-block:: python
                        'data_aug_op_config': {
                            'src': 'aug_src',
                            'tgt': 'aug_tgt'
                        }
        """
        return {
            "data_aug_op": "",
            "data_aug_op_config": {},
            "augment_pack_names": {},
            "@no_typecheck": [
                "data_aug_op_config",
                "augment_pack_names",
            ],
        }



import itertools
from enum import Enum
from typing import Optional

from forte.data.base_pack import PackType

__all__ = ["ProcessJobStatus", "ProcessJob"]

ProcessJobStatus = Enum("ProcessJobStatus", "UNPROCESSED QUEUED PROCESSED")


class ProcessJob:
    counter = itertools.count(0)

    def __init__(self, pack: Optional[PackType], is_poison: bool):
        self.__pack: Optional[PackType] = pack
        self.__is_poison: bool = is_poison
        self.__status = ProcessJobStatus.UNPROCESSED
        self.__id = next(ProcessJob.counter)

    def set_status(self, status):
        self.__status = status

    @property
    def id(self):
        return self.__id

    @property
    def pack(self) -> PackType:
        if self.__pack is None:
            raise ValueError("This job do not have a valid pack.")
        return self.__pack

    def alter_pack(self, pack: PackType):
        """
        This class alter the pack in this job. This should only be controlled by
        the system itself. One should not call this function without proper
        understanding.

        Args:
            pack: The pack to be used to replace into this job.

        Returns:
            None
        """
        self.__pack = pack

    @property
    def is_poison(self) -> bool:
        return self.__is_poison

    @property
    def status(self):
        return self.__status



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from ft.onto.base_ontology import Sentence, Token
from forte.data import DataPack
from forte.processors.base import PackProcessor

__all__ = ["PeriodSentenceSplitter", "WhiteSpaceTokenizer"]


class PeriodSentenceSplitter(PackProcessor):
    """
    A processor that create sentences based on periods.
    """

    def _process(self, input_pack: DataPack):
        pattern = "\\.\\s*"
        start = 0

        for m in re.finditer(pattern, input_pack.text):
            end = m.end()
            Sentence(input_pack, start, end)
            start = end

        if start < len(input_pack.text):
            input_pack.add_entry(
                Sentence(input_pack, start, len(input_pack.text))
            )


class WhiteSpaceTokenizer(PackProcessor):
    """
    A simple processor that split the tokens based on white space.
    """

    def _process(self, input_pack: DataPack):
        pattern = r"\s+"
        start = 0

        for m in re.finditer(pattern, input_pack.text):
            input_pack.add_entry(Token(input_pack, start, m.start()))
            start = m.end()

        if start < len(input_pack.text):
            input_pack.add_entry(Token(input_pack, start, len(input_pack.text)))



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The reader that reads plain text data into Datapacks.
"""
import os
from typing import Any, Iterator, Dict, Set

from forte.data.data_pack import DataPack
from forte.data.data_utils_io import dataset_path_iterator
from forte.data.base_reader import PackReader
from ft.onto.base_ontology import Document

__all__ = [
    "PlainTextReader",
]


class PlainTextReader(PackReader):
    r""":class:`PlainTextReader` is designed to read in plain text dataset."""

    def _collect(self, text_directory) -> Iterator[Any]:  # type: ignore
        r"""Should be called with param ``text_directory`` which is a path to a
        folder containing txt files.

        Args:
            text_directory: text directory containing the files.

        Returns:
            Iterator over paths to .txt files
        """
        return dataset_path_iterator(text_directory, self.configs.file_ext)

    def _cache_key_function(self, text_file: str) -> str:
        return os.path.basename(text_file)

    # pylint: disable=unused-argument
    def text_replace_operation(self, text: str):
        return []

    def _parse_pack(self, file_path: str) -> Iterator[DataPack]:
        pack = DataPack()

        with open(file_path, "r", encoding="utf8", errors="ignore") as file:
            text = file.read()

        pack.set_text(text, replace_func=self.text_replace_operation)

        Document(pack, 0, len(pack.text))

        pack.pack_name = file_path
        yield pack

    @classmethod
    def default_configs(cls):
        return {"file_ext": ".txt"}

    def record(self, record_meta: Dict[str, Set[str]]):
        r"""Method to add output type record of `PlainTextReader` which is
        `ft.onto.base_ontology.Document` with an empty set
        to :attr:`forte.data.data_pack.Meta.record`.

        Args:
            record_meta: the field in the datapack for type record that need to
                fill in for consistency checking.
        """
        record_meta["ft.onto.base_ontology.Document"] = set()



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
We are going to reuse the configuration class HParams in Texar. However, the
name might be confusing in the context of processors, so we rename them to
Config here.
"""
from typing import Dict
from asyml_utilities.hyperparams import HParams

__all__ = ["Config"]

Config = HParams


def merge_configs(config: Dict, parent_config: Dict) -> Dict:
    return HParams(config, parent_config, True).todict()



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import List, Iterator, Tuple, Any, Optional

__all__ = ["BaseStore"]


class BaseStore:
    r"""The base class which will be used by :class:`~forte.data.data_store.DataStore`."""

    def __init__(self):
        r"""
        This is a base class for the efficient underlying data structure. A
        current implementation of ``BaseStore`` is ``DataStore``.

        A ``BaseStore`` contains a collection of Forte entries.
        Each entry type contains some subtypes, which could have
        various fields stored in entry lists.
        """

    @abstractmethod
    def add_annotation_raw(self, type_name: str, begin: int, end: int) -> int:
        r"""This function adds an annotation entry with ``begin`` and ``end``
        indices to the ``type_name`` sorted list in ``self.__elements``,
        returns the ``tid`` for the inserted entry.

        Args:
            type_name: The index of Annotation sorted list in ``self.__elements``.
            begin: Begin index of the entry.
            end: End index of the entry.
        Returns:
            ``tid`` of the entry.
        """
        raise NotImplementedError

    @abstractmethod
    def add_link_raw(
        self, type_name: str, parent_tid: int, child_tid: int
    ) -> Tuple[int, int]:
        r"""This function adds a link entry with ``parent_tid`` and ``child_tid``
        to the ``type_name`` list in ``self.__elements``, returns the ``tid`` and the
        ``index_id`` for the inserted entry in the list. This ``index_id`` is the
        index of the entry in the ``type_name`` list.

        Args:
            type_name: The index of Link list in ``self.__elements``.
            parent_tid: ``tid`` of the parent entry.
            child_tid: ``tid`` of the child entry.

        Returns:
            ``tid`` of the entry and its index in the ``type_name`` list.

        """
        raise NotImplementedError

    @abstractmethod
    def add_group_raw(
        self, type_name: str, member_type: str
    ) -> Tuple[int, int]:
        r"""This function adds a group entry with ``member_type`` to the
        ``type_name`` list in ``self.__elements``, returns the ``tid`` and the
        ``index_id`` for the inserted entry in the list. This ``index_id`` is the
        index of the entry in the ``type_name`` list.

        Args:
            type_name: The index of Group list in ``self.__elements``.
            member_type: Fully qualified name of its members.

        Returns:
            ``tid`` of the entry and its index in the ``type_name`` list.

        """
        raise NotImplementedError

    @abstractmethod
    def set_attribute(self, tid: int, attr_name: str, attr_value: Any):
        r"""This function locates the entry data with ``tid`` and sets its
        ``attr_name`` with ``attr_value``.

        Args:
            tid: Unique Id of the entry.
            attr_name: Name of the attribute.
            attr_value: Value of the attribute.
        """
        raise NotImplementedError

    @abstractmethod
    def set_attr(self, tid: int, attr_id: int, attr_value: Any):
        r"""This function locates the entry data with ``tid`` and sets its
        attribute ``attr_id``  with value ``attr_value``.
        Called by `set_attribute()`.

        Args:
            tid: Unique id of the entry.
            attr_id: Id of the attribute.
            attr_value: value of the attribute.

        """

        raise NotImplementedError

    @abstractmethod
    def get_attribute(self, tid: int, attr_name: str):
        r"""This function finds the value of ``attr_name`` in entry with
        ``tid``.

        Args:
            tid: Unique id of the entry.
            attr_name: Name of the attribute.

        Returns:
            The value of ``attr_name`` for the entry with ``tid``.
        """
        raise NotImplementedError

    @abstractmethod
    def get_attr(self, tid: int, attr_id: int):
        r"""This function locates the entry data with ``tid`` and gets the value
        of ``attr_id``  of this entry. Called by `get_attribute()`.

        Args:
            tid: Unique id of the entry.
            attr_id: Id of the attribute.

        Returns:
            The value of ``attr_id``  for the entry with ``tid``.
        """

        raise NotImplementedError

    @abstractmethod
    def delete_entry(self, tid: int):
        r"""This function removes the entry with ``tid`` from the data store.

        Args:
            tid: Unique id of the entry.

        """

        raise NotImplementedError

    @abstractmethod
    def get_entry(self, tid: int) -> Tuple[List, str]:
        r"""Look up the entry_dict with key ``tid``. Return the entry and its
        ``type_name``.

        Args:
            tid: Unique id of the entry.

        Returns:
            The entry which ``tid`` corresponds to and its ``type_name``.

        """
        raise NotImplementedError

    @abstractmethod
    def get_entry_index(self, tid: int) -> int:
        r"""Look up the entry_dict with key ``tid``. Return the ``index_id`` of
        the entry.

        Args:
            tid: Unique id of the entry.

        Returns:
            Index of the entry which ``tid`` corresponds to in the
            ``entry_type`` list.

        """
        raise NotImplementedError

    @abstractmethod
    def get(self, type_name: str, include_sub_type: bool) -> Iterator[List]:
        r"""This function fetches entries from the data store of
        type ``type_name``.

        Args:
            type_name: The index of the list in ``self.__elements``.
            include_sub_type: A boolean to indicate whether get its subclass.

        Returns:
            An iterator of the entries matching the provided arguments.

        """

        raise NotImplementedError

    @abstractmethod
    def next_entry(self, tid: int) -> Optional[List]:
        r"""Get the next entry of the same type as the ``tid`` entry.

        Args:
            tid: Unique id of the entry.

        Returns:
            The next entry of the same type as the ``tid`` entry.

        """

        raise NotImplementedError

    @abstractmethod
    def prev_entry(self, tid: int) -> Optional[List]:
        r"""Get the previous entry of the same type as the ``tid`` entry.

        Args:
            tid: Unique id of the entry.

        Returns:
            The previous entry of the same type as the ``tid`` entry.

        """

        raise NotImplementedError



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Tuple, Type, Union

import numpy as np
from forte.data.ontology.core import Entry
from forte.data.span import Span

__all__ = ["ReplaceOperationsType", "DataRequest", "MatrixLike"]

ReplaceOperationsType = List[Tuple[Span, str]]

DataRequest = Dict[Type[Entry], Union[Dict, List]]

try:
    import torch

    MatrixLike = Union[torch.TensorType, np.ndarray, List]
except ImportError:
    MatrixLike = Union[np.ndarray, List]  # type: ignore



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Processors that handle query
"""
from abc import ABC
from typing import Union, Tuple, Dict, Any

import numpy as np

from forte.common import EntryNotFoundError
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack
from forte.data.ontology.top import Query
from forte.processors.base.pack_processor import MultiPackProcessor

__all__ = ["QueryProcessor"]

QueryType = Union[Dict[str, Any], np.ndarray]


class QueryProcessor(MultiPackProcessor, ABC):
    r"""A base class for all processors that handle query creation for
    information retrieval."""

    def _build_query(self, text: str) -> QueryType:
        r"""Subclasses of QueryProcessor need to implement this method to create
        a query from a text string.

        Args:
            text: A string for which a query will be generated

        Returns:
            A str or numpy array representing a query for `text`

        """
        raise NotImplementedError

    def _process_query(
        self, input_pack: MultiPack
    ) -> Tuple[DataPack, QueryType]:
        r"""Subclasses of QueryProcessor should implement this method which
        takes in an `input_pack` and processes it to generate a query.

        Args:
            input_pack: A (data/multi)-pack for which
                a query is generated. If `input_pack` is a multipack, the
                processor should fetch the relevant datapack and process it to
                generate a query.

        Returns:
             A tuple containing the `(query_pack, query)`.

             - If `input_pack` is a multipack, `query_pack` is one of its data
             pack for which a `query` is generated.
             - If `input_pack` is a datapack, `query_pack` is the `input_pack`.
        """
        raise NotImplementedError

    def _process(self, input_pack: MultiPack):
        query_pack, query_value = self._process_query(input_pack)

        # Make sure we only have one query on this pack.
        query: Query
        try:
            query = query_pack.get_single(Query)  # type: ignore
        except EntryNotFoundError:
            query = Query(pack=query_pack)

        query.value = query_value



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The reader that reads html data into Datapacks.
"""
from html.parser import HTMLParser
from html import unescape
import os
import re
from typing import Iterator

from forte.data.span import Span
from forte.data.data_pack import DataPack
from forte.data.data_utils_io import dataset_path_iterator
from forte.data.base_reader import PackReader
from ft.onto.base_ontology import Document

# Regular expressions used for parsing. Borrowed from
# https://github.com/python/cpython/blob/3.6/Lib/html/parser.py

interesting_normal = re.compile("[&<]")
incomplete = re.compile("&[a-zA-Z#]")

entityref = re.compile("&([a-zA-Z][-.a-zA-Z0-9]*)[^a-zA-Z0-9]")
charref = re.compile("&#(?:[0-9]+|[xX][0-9a-fA-F]+)[^0-9a-fA-F]")

starttagopen = re.compile("<[a-zA-Z]")
piclose = re.compile(">")
commentclose = re.compile(r"--\s*>")
# Note:
#  1) if you change tagfind/attrfind remember to update locatestarttagend too;
#  2) if you change tagfind/attrfind and/or locatestarttagend the parser will
#     explode, so don't do it.
# see http://www.w3.org/TR/html5/tokenization.html#tag-open-state
# and http://www.w3.org/TR/html5/tokenization.html#tag-name-state
tagfind_tolerant = re.compile(r"([a-zA-Z][^\t\n\r\f />\x00]*)(?:\s|/(?!>))*")
attrfind_tolerant = re.compile(
    r'((?<=[\'"\s/])[^\s/>][^\s/=>]*)(\s*=+\s*'
    r'(\'[^\']*\'|"[^"]*"|(?![\'"])[^>\s]*))?(?:\s|/(?!>))*'
)
locatestarttagend_tolerant = re.compile(
    r"""
  <[a-zA-Z][^\t\n\r\f />\x00]*       # tag name
  (?:[\s/]*                          # optional whitespace before attribute name
    (?:(?<=['"\s/])[^\s/>][^\s/=>]*  # attribute name
      (?:\s*=+\s*                    # value indicator
        (?:'[^']*'                   # LITA-enclosed value
          |"[^"]*"                   # LIT-enclosed value
          |(?!['"])[^>\s]*           # bare value
         )
         (?:\s*,)*                   # possibly followed by a comma
       )?(?:\s|/(?!>))*
     )*
   )?
  \s*                                # trailing whitespace
""",
    re.VERBOSE,
)
endendtag = re.compile(">")
# the HTML 5 spec, section 8.1.2.2, doesn't allow spaces between
# </ and the tag name, so maybe this should be fixed
endtagfind = re.compile(r"</\s*([a-zA-Z][-.a-zA-Z0-9:_]*)\s*>")

__all__ = [
    "HTMLReader",
]


class ForteHTMLParser(HTMLParser):
    r"""Parser that stores spans that HTMLReader can use."""

    def __init__(self):
        super().__init__()
        self.spans = []

    def collect_span(self, begin, end):
        self.spans.append((Span(begin, end), ""))

    # We override the original goahead method and collect the information
    # we need to successfully remove tag information and retrieve the original
    # html document without any loss.
    def goahead(self, end):
        rawdata = self.rawdata
        i = 0
        n = len(rawdata)
        while i < n:
            if self.convert_charrefs and not self.cdata_elem:
                j = rawdata.find("<", i)
                if j < 0:
                    # if we can't find the next <, either we are at the end
                    # or there's more text incoming.  If the latter is True,
                    # we can't pass the text to handle_data in case we have
                    # a charref cut in half at end.  Try to determine if
                    # this is the case before proceeding by looking for an
                    # & near the end and see if it's followed by a space or ;.
                    amppos = rawdata.rfind("&", max(i, n - 34))
                    if amppos >= 0 and not re.compile(r"[\s;]").search(
                        rawdata, amppos
                    ):
                        break  # wait till we get all the text
                    j = n
            else:
                match = self.interesting.search(rawdata, i)  # < or &
                if match:
                    j = match.start()
                else:
                    if self.cdata_elem:
                        break
                    j = n
            if i < j:
                if self.convert_charrefs and not self.cdata_elem:
                    self.handle_data(unescape(rawdata[i:j]))
                else:
                    self.handle_data(rawdata[i:j])
            i = self.updatepos(i, j)
            if i == n:
                break
            startswith = rawdata.startswith
            if startswith("<", i):
                if starttagopen.match(rawdata, i):  # < + letter
                    k = self.parse_starttag(i)
                    self.collect_span(i, k)
                elif startswith("</", i):
                    k = self.parse_endtag(i)
                    self.collect_span(i, k)
                elif startswith("<!--", i):
                    k = self.parse_comment(i)
                    self.collect_span(i, k)
                elif startswith("<?", i):
                    k = self.parse_pi(i)
                    self.collect_span(i, k)
                elif startswith("<!", i):
                    k = self.parse_html_declaration(i)
                    self.collect_span(i, k)
                elif (i + 1) < n:
                    self.handle_data("<")
                    k = i + 1
                else:
                    break
                if k < 0:
                    if not end:
                        break
                    k = rawdata.find(">", i + 1)
                    if k < 0:
                        k = rawdata.find("<", i + 1)
                        if k < 0:
                            k = i + 1
                    else:
                        k += 1
                    if self.convert_charrefs and not self.cdata_elem:
                        self.handle_data(unescape(rawdata[i:k]))
                    else:
                        self.handle_data(rawdata[i:k])
                i = self.updatepos(i, k)
            elif startswith("&#", i):
                match = charref.match(rawdata, i)
                if match:
                    name = match.group()[2:-1]
                    self.handle_charref(name)
                    k = match.end()
                    if not startswith(";", k - 1):
                        k = k - 1
                    i = self.updatepos(i, k)
                else:
                    if ";" in rawdata[i:]:  # bail by consuming &#
                        self.handle_data(rawdata[i : i + 2])
                        i = self.updatepos(i, i + 2)
                    break
            elif startswith("&", i):
                match = entityref.match(rawdata, i)
                if match:
                    name = match.group(1)
                    self.handle_entityref(name)
                    k = match.end()
                    if not startswith(";", k - 1):
                        k = k - 1
                    i = self.updatepos(i, k)
                    continue
                match = incomplete.match(rawdata, i)
                if match:
                    # match.group() will contain at least 2 chars
                    if end and match.group() == rawdata[i:]:
                        k = match.end()
                        if k <= i:
                            k = n
                        i = self.updatepos(i, i + 1)
                    # incomplete
                elif (i + 1) < n:
                    # not the end of the buffer, and can't be confused
                    # with some other construct
                    self.handle_data("&")
                    i = self.updatepos(i, i + 1)
                else:
                    break
            else:
                assert 0, "interesting.search() lied"
        # end while
        if end and i < n and not self.cdata_elem:
            if self.convert_charrefs and not self.cdata_elem:
                self.handle_data(unescape(rawdata[i:n]))
            else:
                self.handle_data(rawdata[i:n])
            i = self.updatepos(i, n)
        # pylint: disable=attribute-defined-outside-init
        self.rawdata = rawdata[i:]


class HTMLReader(PackReader):
    r""":class:`HTMLReader` is designed to read in list of html strings.

    It takes in list of html strings, cleans the HTML tags and stores the
    cleaned text in pack.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.init_with_fileloc = False
        self.init_with_html = False

    def _collect(self, content) -> Iterator[str]:  # type: ignore
        r"""Could be called with a directory, a particular file location or a
        list of strings. If the string is an HTML string, it will be cleaned.

        Args:
            content: either a string, or list of string

        Returns:
            Iterator over the content based on type of input
        """
        if isinstance(content, str):
            # Check if directory
            if os.path.isdir(content):
                self.init_with_fileloc = True
                # TODO: maybe extend it to .txt also if need be?
                return dataset_path_iterator(content, ".html")
            # If file path to a single file, just return the filepath
            elif os.path.isfile(content):

                def data_yielder(data):
                    yield data

                self.init_with_fileloc = True
                return data_yielder(content)
            else:  # Treat it as a string
                content = [content]

        if isinstance(content, list):  # Must be a list of strings now
            self.init_with_html = True

            def data_iterator(data):
                for html_string in data:
                    yield html_string

            return data_iterator(content)

        else:
            raise TypeError(
                f"HTMLReader supports only strings and list of"
                f" strings, Please make sure your inputs are"
                f" correct!"
                f"Found {type(content)} instead!"
            )

    def _parse_pack(self, data_source: str) -> Iterator[DataPack]:
        r"""Takes a string which could be either a filepath or html content and
        converts into a DataPack.

        Args:
            data_source: str that contains text of a document or a filepath

        Returns: DataPack containing Document.
        """
        pack = DataPack()

        # Check if data_source is a filepath
        if self.init_with_fileloc:
            with open(
                data_source, "r", encoding="utf8", errors="ignore"
            ) as file:
                text = file.read()
        # else, must be a string with actual data
        else:
            text = data_source

        self.set_text(pack, text)
        # Note that pack.text can be different from the text passed in, due to
        # the text_replace_operation
        Document(pack, 0, len(pack.text))

        yield pack

    def text_replace_operation(self, text: str):
        r"""Replace html tag locations with blank string.

        Args:
            text: The original html text to be cleaned.

        Returns:
            List[Tuple[Span, str]]: the replacement operations

        """
        parser = ForteHTMLParser()
        parser.feed(text)

        return parser.spans

    def _cache_key_function(self, collection):
        # check if collection is file or html string
        if self.init_with_fileloc:
            return os.path.basename(collection)
        # If html string
        else:
            return str(hash(collection)) + ".html"



# Copyright 2022 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
SkeletonOp is the most low level Op used for Data Augmentation.
Through this Op, users can use in-built utility functions to implement
their own augmentation logic.
"""
from collections import defaultdict
from copy import deepcopy
from typing import (
    DefaultDict,
    Iterable,
    List,
    Tuple,
    Union,
    Dict,
    Any,
    Set,
    cast,
)
from abc import abstractmethod
from bisect import bisect_left, bisect_right
from sortedcontainers.sorteddict import SortedDict
from sortedcontainers.sortedlist import SortedList
from forte.data.data_pack import DataPack
from forte.common.configuration import Config
from forte.data.span import Span
from forte.data.ontology.top import (
    Link,
    Group,
    Annotation,
)
from forte.data.ontology.core import Entry
from forte.common.configurable import Configurable
from forte.utils.utils import create_class_with_kwargs, get_class


__all__ = ["BaseDataAugmentationOp"]


class BaseDataAugmentationOp(Configurable):
    r"""
    The SkeletonOp is the most basic augmentation Op that
    gives users the most amount of freedom to implement their
    logic of augmentation. The users are expected to use the
    provided utility functions to implement their own augmentation
    logic which will then be substantiated into new data packs. This
    Op requires the users to have a relatively stronger understanding of
    Forte's internal setup.
    """

    def __init__(self, configs: Union[Config, Dict[str, Any]]) -> None:

        # :attr:`_replaced_annos`: {datapack id: SortedList[span, new text]}
        # It records the spans replaced by new texts.
        # It is a map from datapack id to a list of tuples
        # (span, new text) inserted by :func:`replace`.
        # The records stored in `_replaced_annos` will then be used to
        # create a new data pack with updated text.

        self._replaced_annos: DefaultDict[
            int, SortedList[Tuple[Span, str]]
        ] = defaultdict(lambda: SortedList([], key=lambda x: x[0]))

        # :attr:`_inserted_text`: {datapack id: Dict{position: length}}
        # It records the position and length of inserted spans,
        # mapping from datapack id to a dictionary (position -> length)
        # inserted by :func:`insert`.
        # The position is the index in the original datapack
        # of insertion, and the length is the length of the inserted string.
        self._inserted_text: DefaultDict[
            int, SortedDict[int, Tuple[int, str]]
        ] = defaultdict(lambda: SortedDict({}))

        # :attr:`_deleted_annos_id`: {datapack id: Set[annotation tid]}
        # It records the deleted span ids, mapping from datapack id
        # to a set of annotation tids appended by :func:`delete`.
        self._deleted_annos_id: DefaultDict[int, Set[int]] = defaultdict(set)

        # :attr:`_data_pack_map`: {orig pack id: new pack id}
        # It maintains a mapping from the pack id
        # of the original pack to the pack id of augmented pack.
        # It is used when copying the MultiPackLink and MultiPackGroup.
        self._data_pack_map: Dict[int, int] = {}

        # :attr:`_entry_maps`: {datapack id: Dict{orig tid, new tid}}
        # It is a map for tracking the annotation ids
        # before and after the auto align. It maps the
        # original annotation tid to the new annotation tid.
        # It is used when copying the Link/Group/MultiPackLink/MultiPackGroup.
        self._entry_maps: Dict[int, Dict[int, int]] = {}

        self.configs = self.make_configs(configs)

    def perform_augmentation(self, input_pack: DataPack) -> DataPack:
        r"""
        Function to apply the defined augmentation function and
        instantiating it into a new data pack. This data pack is
        then returned.

        Args:
            data_pack: The Datapack holding the replaced annotations.
        Returns:
            A new data pack holds the text after replacement.
        """
        self.augment(input_pack)
        augmented_data_pack = self._apply_augmentations(input_pack)
        return augmented_data_pack

    def _copy_link_or_group(
        self,
        entry: Union[Link, Group],
        entry_map: Dict[int, int],
        new_pack: DataPack,
    ) -> bool:
        r"""
        This function copies a Link/Group in the data pack.
        If the children Link/Group does not exist, it will recursively
        create the children Link/Group. If the children Annotation
        does not exist, it will abort and return False.
        Args:
            entry: The Link/Group in the original data pack to copy.
            entry_map: The dictionary mapping original entry to copied entry.
            new_pack: The new data pack, which is the destination of copy.
        Returns:
            A bool value indicating whether the copy happens successfully.
        """

        # If the entry has been copied, return True.
        if entry.tid in entry_map:
            return True

        # The entry should be either Link or Group.
        is_link: bool = isinstance(entry, Link)

        # Get the children entries.
        children: List[Entry]
        if is_link:
            children = [entry.get_parent(), entry.get_child()]  # type: ignore
        else:
            children = entry.get_members()  # type: ignore

        # Copy the children entries.
        new_children: List[Entry] = []
        for child_entry in children:
            if isinstance(child_entry, (Link, Group)):
                # Recursively copy the children Links/Groups.
                if not self._copy_link_or_group(
                    child_entry, entry_map, new_pack
                ):
                    return False
            else:
                # Children Annotation must have been copied.
                if child_entry.tid not in entry_map:
                    return False
            new_child: Entry = new_pack.get_entry(entry_map[child_entry.tid])
            new_children.append(new_child)

        # Create the new entry and add to the new pack.
        new_entry: Entry
        if is_link:
            entry = cast(Link, entry)
            new_link_parent: Entry = new_children[0]
            new_link_child: Entry = new_children[1]
            new_entry = type(entry)(
                new_pack, new_link_parent, new_link_child  # type: ignore
            )
        else:
            entry = cast(Group, entry)
            new_entry = type(entry)(new_pack, new_children)  # type: ignore
        new_pack.add_entry(new_entry)
        entry_map[entry.tid] = new_entry.tid
        return True

    def _overlap_with_existing(self, pid: int, begin: int, end: int) -> bool:
        r"""
        This function will check whether the new span
        has an overlap with any existing spans.

        Args:
            pid: Datapack Id.
            begin: The span begin index.
            end: The span end index.
        Returns:
            True if the input span overlaps with
            any existing spans, False otherwise.
        """
        if len(self._replaced_annos[pid]) == 0:
            return False
        ind: int = (
            bisect_left(self._replaced_annos[pid], (Span(begin, begin), "")) - 1
        )
        if ind < 0:
            ind += 1

        while ind < len(self._replaced_annos[pid]):
            span: Span = self._replaced_annos[pid][ind][0]
            if not (span.begin >= end or span.end <= begin):
                return True
            if span.begin > end:
                break
            ind += 1

        return False

    def modify_index(
        self,
        index: int,
        # Both of the following spans should be SortedList.
        # Use List to avoid typing errors.
        old_spans: List[Span],
        new_spans: List[Span],
        is_begin: bool,
        is_inclusive: bool,
    ) -> int:

        r"""
        A helper function to map an index before replacement
        to the index after replacement.
        An index is the character offset in the data pack.
        The `old_spans` are the inputs of replacement, and the new_spans
        are the outputs. Each of the span has start and end index.
        The `old_spans` and `new_spans` are anchors for the mapping,
        because we depend on them to determine the position change of the
        index.
        Given an index, the function will find its the nearest
        among the old spans before the index, and calculate the difference
        between the position of the old span and its corresponding new span.
        The position change is then applied to the input index. An updated
        index is then calculated and returned.
        An inserted span might be included as a part of another span.
        For example, given a sentence "I love NLP.", if we insert a
        token "Yeah" at the beginning of the sentence(index=0), the Sentence
        should include the new Token, i.e., the Sentence will have a
        start index equals to 0. In this case, the parameter is_inclusive
        should be True. However, for another Token "I", it should not include
        the new token, so its start index will be larger than 0.
        The parameter in_inclusive should be False.
        The input index could be the start or end index of a span, i.e., the
        left or right boundary of the span. If there is an insertion in the span,
        we should treat the two boundaries in different ways. For example,
        we have a paragraph with two sentences "I love NLP! You love NLP too."
        If we append another "!" to the end of the first sentence, when modifying
        the end index of the first Sentence, it should be pushed right to include
        the extra exclamation. In this case, the is_begin is False. However, if
        we prepend an "And" to the second sentence, when modifying the start index
        of the second Sentence, it should be pushed left to include the new Token.
        In this case, the `is_begin` is True.

        Args:
            index: The index to map.
            old_spans: The spans before replacement. It should be
                a sorted list in ascending order.
            new_spans: The spans after replacement. It should be
                a sorted list in ascending order.
            is_begin: True if the input index is the start index of a span.
            is_inclusive: True if the span constructed by the aligned
                index should include inserted spans.

        Returns:
            The aligned index.

        If the old spans are [0, 1], [2, 3], [4, 6],
        the new spans are [0, 4], [5, 7], [8, 11],
        the input index is 3, and there are no insertions,
        the algorithm will first locate the last span with
        a begin index less or equal than the target index,
        ([2,3]), and find the corresponding span in new spans([5,7]).
        Then we calculate the delta index(7-3=4) and update our
        input index(3+4=7). The output then is 7.

        Note that when the input index locates inside the old spans,
        instead of on the boundary of the spans, we compute the return
        index so that it maintains the same offset to the begin of the
        span it belongs to. In the above example, if we change the input
        index from 3 to 5, the output will become 9, because we locates
        the input index in the third span [4, 6] and use the same offset
        5-4=1 to calculate the output 8+1=9.

        When insertion is considered, there will be spans
        with the same begin index, for example,
        [0, 1], [1, 1], [1, 2]. The span [1, 1] indicates an insertion
        at index 1, because the insertion can be considered as a
        replacement of an empty input span, with a length of 0.
        The output will be affected by whether to include the inserted
        span(is_inclusive), and whether the input index is a begin or
        end index of its span(is_begin).

        If the old spans are [0, 1], [1, 1], [1, 2],
        the new spans are [0, 2], [2, 4], [4, 5],
        the input index is 1, the output will be 2 if both
        is_inclusive and is_begin are True,
        because the inserted [1, 1] should be included in the span.
        If the `is_inclusive=True` but `is_begin=False`, the output will be
        4 because the index is an end index of the span.
        """

        # Get the max index for binary search.
        max_index: int = old_spans[-1].end + 1
        max_index = max(max_index, index)

        # This is the last span that has a start index less than
        # the input index. The position change of this span determines
        # the modification we will apply to the input index.
        last_span_ind: int = bisect_right(old_spans, Span(index, max_index)) - 1

        # If there is an inserted span, it will always be the first of
        # those spans with the same begin index. For example, given spans
        # [1, 1], [1, 2], The inserted span [1, 1] will be in the front of
        # replaced span [1, 2], because it has the smallest end index.
        if last_span_ind >= 0:
            if is_inclusive:
                if is_begin:
                    # When inclusive, move the begin index
                    # to the left to include the inserted span.
                    if (
                        last_span_ind > 0
                        and old_spans[last_span_ind - 1].begin == index
                    ):
                        # Old spans: [0, 1], [1, 1], [1, 3]
                        # Target index: 1
                        # Change last_span_index from 2 to 1
                        # to include the [1, 1] span.
                        last_span_ind -= 1
                    else:
                        # Old spans: [0, 1], [1, 1], [2, 3]
                        # Target index: 1
                        # last_span_index: 1
                        # No need to change.
                        pass

            else:
                if not is_begin:
                    # When exclusive, move the end index
                    # to the left to exclude the inserted span.
                    if (
                        last_span_ind > 0
                        and old_spans[last_span_ind - 1].begin == index
                    ):
                        # Old spans: [0, 1], [1, 1], [1, 3]
                        # Target index: 1
                        # Change last_span_index from 2 to 0
                        # to exclude the [1, 1] span.
                        last_span_ind -= 2
                    elif (
                        old_spans[last_span_ind].begin == index
                        and old_spans[last_span_ind].end == index
                    ):
                        # Old spans: [0, 1], [1, 1], [2, 3]
                        # Target index: 1
                        # Change last_span_index from 1 to 0
                        # to exclude the [1, 1] span.
                        last_span_ind -= 1

        if last_span_ind < 0:
            # There is no replacement before this index.
            return index
        # Find the nearest anchor point on the left of current index.
        # Start from the span's begin index.
        delta_index: int = (
            new_spans[last_span_ind].begin - old_spans[last_span_ind].begin
        )

        if (
            old_spans[last_span_ind].begin == old_spans[last_span_ind].end
            and old_spans[last_span_ind].begin == index
            and is_begin
            and is_inclusive
        ):
            return index + delta_index

        if new_spans[last_span_ind].begin == new_spans[last_span_ind].end and (
            old_spans[last_span_ind].begin
            <= index
            <= old_spans[last_span_ind].end
        ):
            return new_spans[last_span_ind].begin

        if old_spans[last_span_ind].end <= index:
            # Use the span's end index as anchor, if possible.
            delta_index = (
                new_spans[last_span_ind].end - old_spans[last_span_ind].end
            )
        return index + delta_index

    def _apply_augmentations(
        self,
        data_pack: DataPack,
    ) -> DataPack:

        r"""
        The objective of this function is to actualize the augmentations
        proposed by the augment function. It will copy and update the text
        of datapack and auto-align the annotation spans. The links are also
        copied if its parent & child are both present in the new pack.
        The groups are copied if all its members are present
        in the new pack.

        Args:
            data_pack: The Datapack holding the replaced annotations.

        Returns:
            A new data_pack holds the text after replacement. The annotations
            in the original data pack will be copied and auto-aligned as
            instructed by the "other_entry_policy" in the configuration.
            The links and groups will be copied if their members are copied.
            New annotations added by the `insert_annotated_spans` function
            will also be added to the newly created data pack. Conversely, if
            annotation is deleted by the `delete_annotation` function or an annotation
            exists within a span that is deleted by the `delete_span` function, it will
            not be added to the new data pack.
        """

        replaced_annotations = self._replaced_annos[data_pack.pack_id]

        if len(replaced_annotations) == 0:
            return deepcopy(data_pack)

        spans: List[Span] = [span for span, _ in replaced_annotations]
        replacement_strs: List[str] = [
            replacement_str for _, replacement_str in replaced_annotations
        ]

        # Get the new text for the new data pack.
        new_text: str = ""
        for i, span in enumerate(spans):
            new_span_str = replacement_strs[i]
            # First, get the gap text between last and this span.
            last_span_end: int = spans[i - 1].end if i > 0 else 0
            gap_text: str = data_pack.text[last_span_end : span.begin]
            new_text += gap_text
            # Then, append the replaced new text.
            new_text += new_span_str

        # Finally, append to new_text the text after the last span.
        new_text += data_pack.text[spans[-1].end :]

        # Get the span (begin, end) before and after replacement.
        new_spans: List[Span] = []

        # Bias is the delta between the beginning
        # indices before & after replacement.
        bias: int = 0
        for i, span in enumerate(spans):
            old_begin: int = spans[i].begin
            old_end: int = spans[i].end
            new_begin: int = old_begin + bias
            new_end = new_begin + len(replacement_strs[i])
            new_spans.append(Span(new_begin, new_end))
            bias = new_end - old_end

        new_pack: DataPack = DataPack()
        new_pack.set_text(new_text)

        entry_map: Dict[int, int] = {}
        insert_ind: int = 0
        pid: int = data_pack.pack_id

        # Only iterate over those entries that are necessary. ie. the
        # ones that are inserted or are present in the other_entry_policy
        # config.
        existing_entries = self.configs["other_entry_policy"].keys()

        new_entries: Dict[str, List[Tuple[int, int]]] = {}
        for pos, data in self._inserted_text[pid].items():
            new_entries[data[1]] = new_entries.get(data[1], []) + [
                (pos, data[0])
            ]

        entries_to_copy: Set[str] = set(
            list(existing_entries)
            + [val for val in new_entries if val is not None]
        )

        def _insert_new_span(
            entry_class: str,
            insert_ind: int,
            inserted_annos: List[Tuple[int, int]],
            new_pack: DataPack,
            spans: List[Span],
            new_spans: List[Span],
        ):
            """
            An internal helper function for insertion.

            Args:
                entry_class: The new annotation type to be created.
                insert_ind: The index to be insert.
                inserted_annos: The annotation span information to be inserted.
                new_pack: The new data pack to insert the annotation.
                spans: The original spans before replacement, should be
                  a sorted ascending list.
                new_spans: The original spans before replacement, should be
                  a sorted ascending list.
            """
            pos: int
            length: int
            pos, length = inserted_annos[insert_ind]
            if entry_class is None:
                return
            insert_end: int = self.modify_index(
                pos,
                spans,
                new_spans,
                is_begin=False,
                # Include the inserted span itself.
                is_inclusive=True,
            )
            insert_begin: int = insert_end - length
            new_anno = create_class_with_kwargs(
                entry_class,
                {"pack": new_pack, "begin": insert_begin, "end": insert_end},
            )
            new_pack.add_entry(new_anno)

        # Iterate over all the original entries and modify their spans.
        for entry_to_copy in entries_to_copy:

            class_to_copy = get_class(entry_to_copy)
            insert_ind = 0

            if not issubclass(class_to_copy, Annotation):
                raise AttributeError(
                    f"The entry type to copy from [{entry_to_copy}] is not "
                    f"a sub-class of 'forte.data.ontology.top.Annotation'."
                )

            if entry_to_copy not in new_entries:
                new_entries[entry_to_copy] = []

            orig_annos: Iterable[Annotation] = data_pack.get(class_to_copy)

            for orig_anno in orig_annos:
                # Dealing with insertion/deletion only for augment_entry.
                while (
                    insert_ind < len(new_entries[entry_to_copy])
                    and new_entries[entry_to_copy][insert_ind][0]
                    <= orig_anno.begin
                ):
                    # Preserve the order of the spans with merging sort.
                    # It is a 2-way merging from the inserted spans
                    # and original spans based on the begin index.
                    _insert_new_span(
                        entry_to_copy,
                        insert_ind,
                        new_entries[entry_to_copy],
                        new_pack,
                        spans,
                        new_spans,
                    )
                    insert_ind += 1

                # Deletion
                if orig_anno.tid in self._deleted_annos_id[pid]:
                    continue

                # Auto align the spans.
                span_new_begin: int = orig_anno.begin
                span_new_end: int = orig_anno.end

                if (entry_to_copy in new_entries) or self.configs[
                    "other_entry_policy"
                ][entry_to_copy] == "auto_align":
                    # Only inclusive when the entry is not augmented.
                    # E.g.: A Sentence include the inserted Token on the edge.
                    # E.g.: A Token shouldn't include a nearby inserted Token.
                    is_inclusive = (entry_to_copy in existing_entries) and (
                        entry_to_copy not in new_entries
                    )
                    span_new_begin = self.modify_index(
                        orig_anno.begin, spans, new_spans, True, is_inclusive
                    )
                    span_new_end = self.modify_index(
                        orig_anno.end, spans, new_spans, False, is_inclusive
                    )

                # If an annotation is within a deleted span, there
                # is no need to add that annotation to the new data pack
                if span_new_begin == span_new_end:
                    continue

                new_anno = create_class_with_kwargs(
                    entry_to_copy,
                    {
                        "pack": new_pack,
                        "begin": span_new_begin,
                        "end": span_new_end,
                    },
                )
                new_pack.add_entry(new_anno)
                entry_map[orig_anno.tid] = new_anno.tid

            # Deal with spans after the last annotation in the original pack.
            while insert_ind < len(new_entries[entry_to_copy]):
                _insert_new_span(
                    entry_to_copy,
                    insert_ind,
                    new_entries[entry_to_copy],
                    new_pack,
                    spans,
                    new_spans,
                )
                insert_ind += 1

        # Iterate over and copy the links/groups in the datapack.
        for link in data_pack.get(Link):
            self._copy_link_or_group(link, entry_map, new_pack)
        for group in data_pack.get(Group):
            self._copy_link_or_group(group, entry_map, new_pack)

        self._data_pack_map[pid] = new_pack.pack_id
        self._entry_maps[pid] = entry_map
        return new_pack

    def insert_span(
        self, inserted_text: str, data_pack: DataPack, pos: int
    ) -> bool:
        r"""
        This is a utility function to insert a new text span to a data pack.
        The inserted span will not have any annotation associated with it.
        After getting the inserted text, it will register the input & output
        for later batch process of building the new data pack.
        The insertion at each position can only occur once. If there
        is already an insertion at current position, it will abort the
        insertion and return False.

        Args:
            inserted_text: The text string to insert.
            data_pack: The datapack for insertion.
            pos: The position(index) of insertion.
        Returns:
            A bool value. True if the insertion happened, False otherwise.
        """

        pid: int = data_pack.pack_id
        if self._overlap_with_existing(pid, pos, pos):
            return False

        if pos not in self._inserted_text[pid]:
            self._replaced_annos[pid].add((Span(pos, pos), inserted_text))
            self._inserted_text[pid][pos] = (len(inserted_text), None)
            return True
        return False

    def insert_annotated_span(
        self,
        inserted_text: str,
        data_pack: DataPack,
        pos: int,
        annotation_type: str,
    ) -> bool:
        r"""
        This is a utility function to insert a new annotation to a data pack.
        After getting the inserted text, it will register the input & output
        for later batch process of building the new data pack.
        The insertion at each position can only occur once. If there
        is already an insertion at current position, it will abort the
        insertion and return False.

        Args:
            inserted_text: The text string to insert.
            data_pack: The datapack for insertion.
            pos: The position(index) of insertion.
            annotation_type: The type of annotation this span represents.
        Returns:
            A bool value. True if the insertion happened, False otherwise.

        """
        pid: int = data_pack.pack_id
        if self._overlap_with_existing(pid, pos, pos):
            return False

        if pos not in self._inserted_text[pid]:
            self._replaced_annos[pid].add((Span(pos, pos), inserted_text))
            self._inserted_text[pid][pos] = (
                len(inserted_text),
                annotation_type,
            )
            return True
        return False

    def delete_annotation(self, input_anno: Annotation) -> bool:
        r"""
        This is a utility function to delete an annotation. If the same
        annotation is tried to be deleted twice, the function will terminate
        and return False.

        Args:
            input_anno: The annotation to remove.
        Returns:
            A bool value. True if the deletion happened, False otherwise.
        """
        pid: int = input_anno.pack.pack_id
        if self._overlap_with_existing(pid, input_anno.begin, input_anno.end):
            return False

        self._replaced_annos[pid].add((input_anno.span, ""))
        self._deleted_annos_id[pid].add(input_anno.tid)
        return True

    def delete_span(self, data_pack: DataPack, begin: int, end: int) -> bool:
        r"""
        This is a utility function to delete a span of text. If the same
        annotation is tried to be deleted twice, the function will terminate
        and return False. If the method deletes only a portion of an existing annotation,
        The annotation will be calibrated to represent the remaining part of the span. Moreover,
        if the deleted span covers an entire annotation, the entire annotation will be deleted.

        Args:
            input_anno: The annotation to remove.
            begin: The starting position of the span to delete
            end: The ending position of the span to delete
        Returns:
            A bool value. True if the deletion happened, False otherwise.
        """

        pid: int = data_pack.pack_id
        if self._overlap_with_existing(pid, begin, end):
            return False

        self._replaced_annos[pid].add((Span(begin, end), ""))
        return True

    def replace_annotations(
        self, replacement_anno: Annotation, replaced_text
    ) -> bool:
        r"""
        This is a utility function to record specifically a replacement
        of the text in an annotation. With this function, the text inside
        annotation can be replaced with another text. If the same annotation
        is tried to be replaced twice, the function will terminate and
        return False.

        Args:
            input_anno: The annotation to replace.
        Returns:
            A bool value. True if the replacement happened, False otherwise.
        """
        pid: int = replacement_anno.pack.pack_id
        if self._overlap_with_existing(
            pid, replacement_anno.begin, replacement_anno.end
        ):
            return False

        self._replaced_annos[pid].add(
            (Span(replacement_anno.begin, replacement_anno.end), replaced_text)
        )
        return True

    def clear_states(self):
        r"""
        This function clears the states. It should be
        called after processing a multipack.
        """
        self._replaced_annos.clear()
        self._inserted_text.clear()
        self._deleted_annos_id.clear()
        self._data_pack_map.clear()
        self._entry_maps.clear()

    def get_maps(self) -> Tuple[Dict[int, int], Dict[int, Dict[int, int]]]:
        r"""
        This function simply returns the produced data pack
        and entry maps after augmentation.

        Returns:
            A tuple of two elements. The first element is the data pack
            map (dict) and the second element is the entry maps (dict)
        """
        return self._data_pack_map, self._entry_maps

    @abstractmethod
    def augment(self, data_pack: DataPack) -> bool:
        r"""
        This method is left to be implemented by the user of this Op.
        The user can use any of the given utility functions to perform
        augmentation.

        Args:
            input_anno: the input annotation to be replaced.

        Returns:
            A boolean value indicating if the augmentation
            was successful (True) or unsuccessful (False).
        """
        raise NotImplementedError

    @classmethod
    def default_configs(cls):
        """
        Returns:
            A dictionary with the default config for this processor.
        Following are the keys for this dictionary:
            - other_entry_policy:
                A dict specifying the policies for other entries.
                The key should be a full qualified class name.
                The policy(value of the dict) specifies how to process
                the corresponding entries after replacement.

                If the policy is "auto_align", the span of the entry
                will be automatically modified according to its original
                location. However, some spans might become invalid after
                the augmentation, for example, the tokens within a
                replaced sentence may disappear.

                Annotations not in the "other_entry_policy" will not
                be copied to the new data pack. The Links and Groups
                will be copied as well if the annotations they are
                attached to are copied.
                Example:

                    .. code-block:: python

                        'other_entry_policy': {
                            "ft.onto.base_ontology.Document": "auto_align",
                            "ft.onto.base_ontology.Sentence": "auto_align",
                        }
        """
        return {
            "other_entry_policy": {
                "ft.onto.base_ontology.Sentence": "auto_align",
                "ft.onto.base_ontology.Token": "auto_align",
            },
            "@no_typecheck": ["other_entry_policy"],
        }



from typing import Any, Dict

from forte.data.data_pack import DataPack
from forte.data.ontology.core import Entry
from forte.processors.base import PackProcessor
from forte.utils import get_class

__all__ = ["AnnotationRemover"]


class AnnotationRemover(PackProcessor):
    def _process(self, input_pack: DataPack):
        for type_name in self.configs.removal_types:
            type_cls = get_class(type_name)

            # Note: we cannot delete during iteration, which may cause
            # un-expected behavior in the iterator.
            t: Entry
            for t in list(input_pack.get(type_cls)):
                input_pack.delete_entry(t)

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        return {"removal_types": []}
